# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Constraint aggregation methods.

Transform a constraint vector into one scalar equivalent or quasi equivalent constraint.
"""

from __future__ import division, unicode_literals

from functools import wraps
from typing import Any, Callable, Optional, Sequence, Union

from numpy import ndarray

from gemseo.algos.aggregation.core import (
    iks_agg,
    iks_agg_jac_v,
    ks_agg,
    ks_agg_jac_v,
    max_agg,
    max_agg_jac_v,
    sum_square_agg,
    sum_square_agg_jac_v,
)
from gemseo.core.function import MDOFunction


def check_constraint_type(
    function_type,  # type: str
):  # type: (...) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]
    """Decorate a function to check whether it is of the expected type.

    Args:
        function_type: The expected function type, ``"ineq"`` or ``"eq"``.

    Returns:
        The decorated function.
    """

    def decorator(
        func,  # type: Callable[[Any], Any]
    ):  # type: (...) -> Callable[[Any], Any]
        """Decorator to check the aggregation function type.

        Args:
            func: The aggregation function.

        Returns:
            The decorated function.
        """

        @wraps(func)
        def function_wrapper(
            *args,  # type: Any
            **kwargs  # type: Any
        ):  # type: (...) -> Any
            """Check that ``func`` has the type `function_type``.

            Args:
                *args: The positional arguments.
                **kwargs: The keyword arguments.

            Raises:
                ValueError: If the type is not correct.

            Returns:
                The return value of ``func``.
            """
            constr = args[0]

            if constr.f_type != function_type:
                msg = (
                    "{} constraint aggregation is only supported"
                    " for func_type {}, got {}"
                ).format(func.__name__, function_type, constr.f_type)
                raise ValueError(msg)

            return func(*args, **kwargs)

        return function_wrapper

    return decorator


@check_constraint_type("eq")
def aggregate_sum_square(
    constr_fct,  # type: MDOFunction
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> MDOFunction
    """Transform a vector of equalities into a sum of squared constraints.

    Args:
        constr_fct: The initial constraint function.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function.
    """

    def compute(x):
        return sum_square_agg(constr_fct(x), indices=indices, scale=scale)

    def compute_jac(x):
        return sum_square_agg_jac_v(
            constr_fct(x), constr_fct.jac(x), indices=indices, scale=scale
        )

    return _create_mdofunc(
        constr_fct,
        compute,
        compute_jac,
        "sum²_{}".format(constr_fct.name),
        "sum({}**2)".format(constr_fct.expr),
        "sum_sq_cstr",
    )


@check_constraint_type("ineq")
def aggregate_max(
    constr_fct,  # type: MDOFunction
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> MDOFunction
    """Transform a vector of equalities into a max of all values.

    Args:
        constr_fct: The initial constraint function.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function.
    """

    def compute(x):
        return max_agg(constr_fct(x), indices=indices, scale=scale)

    def compute_jac(x):
        return max_agg_jac_v(
            constr_fct(x), constr_fct.jac(x), indices=indices, scale=scale
        )

    return _create_mdofunc(
        constr_fct,
        compute,
        compute_jac,
        "max_" + constr_fct.name,
        "max({})".format(constr_fct.expr),
        "max_cstr",
    )


@check_constraint_type("ineq")
def aggregate_iks(
    constr_fct,  # type: MDOFunction
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> MDOFunction
    """Constraints aggregation method for inequality constraints.

    See :cite:t:`kennedy2015improved`.

    Args:
        constr_fct: The initial constraint function.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.
        rho: The multiplicative parameter in the exponential.

    Returns:
        The aggregated function.
    """

    def compute(x):
        return iks_agg(constr_fct(x), indices=indices, rho=rho, scale=scale)

    def compute_jac(x):
        return iks_agg_jac_v(
            constr_fct(x), constr_fct.jac(x), indices=indices, rho=rho, scale=scale
        )

    return _create_mdofunc(
        constr_fct,
        compute,
        compute_jac,
        "IKS({})".format(constr_fct.name),
        "IKS({})".format(constr_fct.expr),
        "IKS",
    )


@check_constraint_type("ineq")
def aggregate_ks(
    constr_fct,  # type: MDOFunction
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> MDOFunction
    """Aggregate constraints for inequality constraints.

    See :cite:t:`kennedy2015improved` and  :cite:t:`kreisselmeier1983application`.

    Args:
        constr_fct: The initial constraint function.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.
        rho: The multiplicative parameter in the exponential.

    Returns:
        The aggregated function.
    """

    def compute(x):
        return ks_agg(constr_fct(x), indices=indices, rho=rho, scale=scale)

    def compute_jac(x):
        return ks_agg_jac_v(
            constr_fct(x), constr_fct.jac(x), indices=indices, rho=rho, scale=scale
        )

    return _create_mdofunc(
        constr_fct,
        compute,
        compute_jac,
        "KS({})".format(constr_fct.name),
        "KS({})".format(constr_fct.expr),
        "KS",
    )


def _create_mdofunc(
    constr_fct,  # type: MDOFunction
    compute_fct,  # type: Callable[[ndarray], ndarray]
    compute_jac_fct,  # type: Callable[[ndarray], ndarray]
    new_name,  # type: str
    new_expr,  # type: str
    new_output_names,  # type: Sequence[str]
):  # type: (...) -> MDOFunction
    """Create an aggregated MDOFunction from a constraint function.

    Args:
        constr_fct: The initial constraint function.
        compute_fct: The aggregated compute function.
        compute_jac_fct: The aggregated compute function jacobian.
        new_name: The name of aggregated function.
        new_expr: The aggregated function expression.
        new_output_names: The aggregated function output names.

    Returns:
        The aggregated MDOFunction.
    """
    return MDOFunction(
        compute_fct,
        new_name,
        constr_fct.f_type,
        compute_jac_fct,
        new_expr,
        constr_fct.args,
        1,
        new_output_names,
        constr_fct.force_real,
    )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A 1D Jameson sensor."""
from __future__ import division, unicode_literals

from numpy import abs as np_abs
from numpy import amax, ndarray

from gemseo.mlearning.transform.transformer import Transformer, TransformerFitOptionType


class JamesonSensor(Transformer):
    """A 1D Jameson Sensor."""

    def __init__(
        self,
        name="JamesonSensor",  # type: str
        threshold=0.3,  # type:float
        removing_part=0.01,  # type:float
        dimension=1,  # type: int
    ):  # type: (...) -> None
        """
        Args:
            name: A name for this transformer.
            threshold: The value to add to the denominator
                to avoid zero division.
            removing_part: The level of the signal to
                remove in order to avoid leading and trailing edge effects.
            dimension: The dimension of the mesh.
        """
        super(JamesonSensor, self).__init__(name)
        self.threshold = threshold
        self.removing_part = removing_part
        self.dimension = dimension

    def fit(
        self,
        data,  # type: ndarray
        *args  # type: TransformerFitOptionType
    ):  # type: (...) -> None
        self.threshold = self.threshold * amax(data)

    def transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        mesh_size = data.shape[1] - 2
        min_mesh_size = int(mesh_size * self.removing_part)
        max_mesh_size = int(mesh_size * (1 - self.removing_part))
        norm = np_abs(data[:, :-2])
        norm += 2 * np_abs(data[:, 1:-1])
        norm += np_abs(data[:, 2:])
        norm = norm + self.threshold
        result = abs(data[:, :-2] - 2 * data[:, 1:-1] + data[:, 2:]) / norm
        result = result[:, min_mesh_size:max_mesh_size]
        return result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Factory base class."""

from __future__ import division, unicode_literals

import importlib
import logging
import os
import pkgutil
import sys
from typing import Any, Dict, Iterable, List, Optional, Type, Union

from gemseo.core.json_grammar import JSONGrammar
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.py23_compat import lru_cache
from gemseo.utils.source_parsing import get_default_options_values, get_options_doc

LOGGER = logging.getLogger(__name__)


@lru_cache(maxsize=None)
class Factory(object):
    """Factory of class objects.

    This factory can create an object from a base class
    or any of its sub-classes
    that can be imported from the given modules sources.
    There are 3 sources of modules that can be searched:

    - fully qualified module names (such as gemseo.problems, ...),
    - the environment variable "GEMSEO_PATH" may contain the list of directories,
    - |g| plugins, i.e. packages whose name starts with ``gemseo_``.

    If a class,
    despite being a sub-class of the base class,
    or even the base class itself,
    does not belong to the modules sources
    then it is not taken into account
    by the factory.
    """

    # Names of the environment variable to search for classes
    __GEMSEO_PATH = "GEMSEO_PATH"
    __GEMS_PATH = "GEMS_PATH"

    # Allowed prefix for naming a plugin importable from sys.path
    __PLUGIN_PREFIX = "gemseo_"

    def __init__(
        self,
        base_class,  # type: Type[Any]
        module_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        # noqa: D205, D212, D405, D415
        """
        Args:
            base_class: The base class to be considered.
            module_names: The fully qualified modules names to be searched.
        """
        if not isinstance(base_class, type):
            raise TypeError("Class to search must be a class!")

        self.__base_class = base_class
        self.__module_names = module_names or []
        self.__names_to_classes = {}
        self.failed_imports = {}

        self.update()

    def update(self):  # type: (...) -> None
        """Search for the classes that can be instantiated.

        The search is done in the following order:
            1. The fully qualified module names
            2. The plugin packages
            3. The packages from the environment variables
        """
        module_names = list(self.__module_names)

        # Import internal packages
        for module_name in module_names:
            self.__import_modules_from(module_name)

        # Import plugins packages
        for _, module_name, _ in pkgutil.iter_modules():
            if module_name.startswith(self.__PLUGIN_PREFIX):
                self.__import_modules_from(module_name)
                module_names += [module_name]

        gems_path = os.environ.get(self.__GEMS_PATH)
        if gems_path is not None:
            msg = (
                "GEMS is now named GEMSEO. "
                "The GEMS_PATH environment variable is now deprecated "
                "and it is strongly recommended "
                "to use the GEMSEO_PATH environment variable "
                "instead to register your GEMSEO plugins."
            )
            LOGGER.warning(msg)

        # Import from environment variable paths
        for env_variable in [self.__GEMSEO_PATH, self.__GEMS_PATH]:
            module_names += self.__import_modules_from_env_var(env_variable)

        names_to_classes = self.__get_sub_classes(self.__base_class)
        for name, cls in names_to_classes.items():
            if self.__is_class_in_modules(module_names, cls):
                self.__names_to_classes[name] = cls

    def __log_import_failure(
        self, pkg_name  # type: str
    ):  # type: (...) -> None
        """Log import failures.

        Args:
            pkg_name: The name of a package that failed to be imported.
        """
        LOGGER.debug("Failed to import package %s", pkg_name)
        self.failed_imports[pkg_name] = ""

    def __import_modules_from_env_var(
        self, env_variable  # type: str
    ):  # type: (...) -> List[str]
        """Import the modules from the path given by an environment variable.

        Args:
            env_variable: The name of an environment variable.

        Returns:
            The imported fully qualified module names.
        """
        g_path = os.environ.get(env_variable)
        if g_path is None:
            return []

        if ":" in g_path:
            paths = g_path.split(":")
        else:
            paths = [g_path]

        # temporary make the paths visible to the import machinery
        for path in paths:
            sys.path.insert(0, path)

        mod_names = list()
        for _, mod_name, _ in pkgutil.iter_modules(path=paths):
            self.__import_modules_from(mod_name)
            mod_names += [mod_name]

        for _ in paths:
            sys.path.pop(0)

        return mod_names

    def __import_modules_from(
        self, pkg_name  # type: str
    ):  # type: (...) -> None
        """Import all the modules from a package.

        Args:
            pkg_name: The name of the package.
        """
        pkg = importlib.import_module(pkg_name)

        if not hasattr(pkg, "__path__"):
            # not a package so no more module to import
            return

        for _, mod_name, _ in pkgutil.walk_packages(
            pkg.__path__, pkg.__name__ + ".", self.__log_import_failure
        ):
            try:
                importlib.import_module(mod_name)
            except Exception as err:  # pylint: disable=(broad-except
                LOGGER.debug("Failed to import module: %s", mod_name, exc_info=True)
                self.failed_imports[mod_name] = err

    def __get_sub_classes(
        self, cls  # type: Type[Any]
    ):  # type: (...) -> Dict[str, Type[Any]]
        """Find all the sub classes of a class.

        The class names are unique,
        the last imported is kept when more than one class have the same name.

        Args:
            cls: A class.

        Returns:
            A mapping from the names to the unique sub-classes.
        """
        all_sub_classes = {}
        for sub_class in cls.__subclasses__():
            sub_classes = {sub_class.__name__: sub_class}
            sub_classes.update(self.__get_sub_classes(sub_class))
            for cls_name, _cls in sub_classes.items():
                all_sub_classes[cls_name] = _cls
        return all_sub_classes

    @staticmethod
    def __is_class_in_modules(
        module_names,  # type: str
        cls,  # type: Type[Any]
    ):  # type: (...) -> bool
        """Return whether a class belongs to given modules.

         Args:
            module_names: The names of the modules.
            cls: The class.

        Returns:
            Whether the class belongs to the modules.
        """
        for name in module_names:
            if cls.__module__.startswith(name):
                return True
        return False

    @property
    @lru_cache()
    def classes(self):  # type: (...) -> List[str]
        """Return the available classes.

        Returns:
            The sorted names of the available classes.
        """
        return sorted(self.__names_to_classes.keys())

    def is_available(
        self, name  # type: str
    ):  # type: (...) -> bool
        """Return whether a class can be instantiated.

        Args:
            name: The name of the class.

        Returns:
            Whether the class can be instantiated.
        """
        return name in self.__names_to_classes

    def get_class(
        self, name  # type: str
    ):  # type: (...) -> Type[Any]
        """Return a class from its name.

        Args:
            name: The name of the class.

        Returns:
            The class.

        Raises:
            ImportError: If the class is not available.
        """
        try:
            return self.__names_to_classes[name]
        except KeyError:
            msg = "Class {} is not available!\nAvailable ones are: {}".format(
                name, ", ".join(sorted(self.__names_to_classes.keys()))
            )
            raise ImportError(msg)

    def create(
        self,
        class_name,  # type: str
        **options  # type: Any
    ):  # type: (...) -> Any
        """Return an instance of a class.

        Args:
            class_name: The name of the class.
            **options: The arguments to be passed to the class constructor.

        Returns:
            The instance of the class.

        Raises:
            TypeError: If the class cannot be instantiated.
        """
        cls = self.get_class(class_name)
        try:
            return cls(**options)
        except TypeError:
            LOGGER.error(
                "Failed to create class %s with arguments %s", class_name, options
            )
            raise

    def get_options_doc(
        self, name  # type: str
    ):  # type: (...) -> Dict[str, str]
        """Return the constructor documentation of a class.

        Args:
            name: The name of the class.

        Returns:
            The mapping from the argument names to their documentation.
        """
        cls = self.get_class(name)
        return get_options_doc(cls.__init__)

    def get_default_options_values(
        self, name  # type: str
    ):  # type: (...) -> Dict[str, Union[str,int,float,bool]]
        """Return the constructor kwargs default values of a class.

        Args:
            name: The name of the class.

        Returns:
            The mapping from the argument names to their default values.
        """
        cls = self.get_class(name)
        return get_default_options_values(cls)

    def get_options_grammar(
        self,
        name,  # type: str
        write_schema=False,  # type: bool
        schema_file=None,  # type: Optional[str]
    ):  # type: (...) -> JSONGrammar
        """Return the options JSON grammar for a class.

        Attempt to generate a JSONGrammar
        from the arguments of the __init__ method of the class.

        Args:
            name: The name of the class.
            write_schema: If True, write the JSON schema to a file.
            schema_file: The path to the JSON schema file.
                If None, the file is saved in the current directory in a file named
                after the name of the class.

        Returns:
            The JSON grammar.
        """
        args_dict = self.get_default_options_values(name)
        opts_doc = self.get_options_doc(name)
        opts_doc = {k: v for k, v in opts_doc.items() if k in args_dict}
        grammar = JSONGrammar(name)

        grammar.initialize_from_base_dict(
            args_dict,
            description_dict=opts_doc,
        )

        if write_schema:
            grammar.write_schema(schema_file)

        # Remove None args from required
        sch_dict = grammar.schema.to_dict()
        required = sch_dict["required"]
        has_changed = False

        for opt, val in args_dict.items():
            if val is None and opt in required:
                required.remove(opt)
                has_changed = True

        if has_changed:
            grammar = JSONGrammar(name, schema=sch_dict)

        return grammar

    def get_sub_options_grammar(
        self,
        name,  # type: str
        **options  # type: str
    ):  # type: (...) -> JSONGrammar
        """Return the JSONGrammar of the sub options of a class.

        Args:
            name: The name of the class.
            **options: The options to be passed to the class required to deduce
                the sub options.

        Returns:
            The JSON grammar.
        """
        cls = self.get_class(name)
        return cls.get_sub_options_grammar(**options)

    def get_default_sub_options_values(
        self,
        name,  # type: str
        **options  # type: str
    ):  # type: (...) -> JSONGrammar
        """Return the default values of the sub options of a class.

        Args:
            name: The name of the class.
            **options: The options to be passed to the class required to deduce
                the sub options.

        Returns:
            The JSON grammar.
        """
        cls = self.get_class(name)
        return cls.get_default_sub_options_values(**options)

    def __str__(self):  # type: (...) -> str
        return "Factory({})".format(self.__base_class.__name__)

    def __repr__(self):  # type: (...) -> str
        # Display the successfully loaded modules and the failed imports with the reason
        table = PrettyTable(
            ["Module", "Is available ?", "Purpose or error message"],
            title=self.__base_class.__name__,
            min_table_width=120,
            max_table_width=120,
        )

        row_dict = {}
        for cls in self.__names_to_classes.values():
            msg = ""
            try:
                msgs = cls.__doc__.split("\n")
                while msgs and msg == "":
                    msg = msgs[0]
                    del msgs[0]
            except Exception:  # pylint: disable=broad-except
                pass

            key = cls.__name__
            row_dict[key] = [key, "Yes", msg]

        for key, err in self.failed_imports.items():
            row_dict[key] = [key, "No", str(err)]

        # Take them all and then sort them for pretty printing
        for key in sorted(row_dict.keys()):
            table.add_row(row_dict[key])

        return table.get_string()



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A chain of MDAs to build hybrids of MDA algorithms sequentially
***************************************************************
"""
from __future__ import division, unicode_literals

from gemseo.core.discipline import MDODiscipline
from gemseo.mda.gauss_seidel import MDAGaussSeidel
from gemseo.mda.mda import MDA
from gemseo.mda.newton import MDANewtonRaphson


class MDASequential(MDA):
    """Perform a MDA defined as a sequence of elementary MDAs."""

    def __init__(
        self,
        disciplines,
        mda_sequence,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        max_mda_iter=10,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param mda_sequence: sequence of MDAs
        :type mda_sequence: list(MDA)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param name: name
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param use_lu_fact: use LU factorization
        :type use_lu_fact: bool
        """
        super(MDASequential, self).__init__(
            disciplines,
            name=name,
            grammar_type=grammar_type,
            max_mda_iter=max_mda_iter,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
        )
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()

        self.mda_sequence = mda_sequence
        for mda in self.mda_sequence:
            mda.reset_history_each_run = True

    def _initialize_grammars(self):
        """Defines all inputs and outputs."""
        for discipline in self.disciplines:
            self.input_grammar.update_from(discipline.input_grammar)
            self.output_grammar.update_from(discipline.output_grammar)

    def _run(self):
        """Runs the MDAs in a sequential way.

        :returns: the local data
        """
        self._couplings_warm_start()
        # execute MDAs in sequence
        if self.reset_history_each_run:
            self.residual_history = []
        for mda_i in self.mda_sequence:
            mda_i.reset_statuses_for_run()
            self.local_data = mda_i.execute(self.local_data)
            self.residual_history += mda_i.residual_history
            if mda_i.normed_residual < self.tolerance:
                break


class GSNewtonMDA(MDASequential):
    """Perform some GaussSeidel iterations and then NewtonRaphson iterations."""

    def __init__(
        self,
        disciplines,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        tolerance=1e-6,
        max_mda_iter=10,
        relax_factor=0.99,
        linear_solver="lgmres",
        max_mda_iter_gs=3,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
        **newton_mda_options
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param name: name
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param relax_factor: relaxation factor
        :type relax_factor: float
        :param linear_solver: type of linear solver to be used to solve
            the Newton problem
        :type linear_solver: str
        :param max_mda_iter_gs: maximum number of iterations of the GaussSeidel
            solver
        :type max_mda_iter_gs: int
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        :param newton_mda_options: options passed to the MDANewtonRaphson
        :type newton_mda_options: dict
        """
        mda_gs = MDAGaussSeidel(disciplines, max_mda_iter=max_mda_iter_gs, name=None)
        mda_gs.tolerance = tolerance
        mda_newton = MDANewtonRaphson(
            disciplines,
            max_mda_iter,
            relax_factor,
            name=None,
            grammar_type=grammar_type,
            linear_solver=linear_solver,
            use_lu_fact=use_lu_fact,
            **newton_mda_options
        )
        sequence = [mda_gs, mda_newton]
        super(GSNewtonMDA, self).__init__(
            disciplines,
            sequence,
            name=name,
            grammar_type=grammar_type,
            max_mda_iter=max_mda_iter,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Antoine Dechaume
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Logging tools
=============
"""

import logging
import types


class MultiLineHandlerMixin(object):
    """Stateless mixin class to override logging handlers behavior."""

    @staticmethod
    def __get_raw_record_message(record):
        """Return the raw message of a log record."""
        return record.msg

    def emit(self, record):
        """Emit one logging message per input record line."""
        # compute the message without the logging prefixes (timestamp, level, ...)
        message = record.getMessage()
        # replace getMessage so the message is not computed again when emitting
        # each line
        record.getMessage = types.MethodType(self.__get_raw_record_message, record)
        # backup old raw message
        old_msg = record.msg
        for line in message.split("\n"):
            record.msg = line
            super(MultiLineHandlerMixin, self).emit(record)
        # restore genuine getMessage and raw message for other handlers
        record.getMessage = types.MethodType(logging.LogRecord.getMessage, record)
        record.msg = old_msg


class MultiLineStreamHandler(MultiLineHandlerMixin, logging.StreamHandler):
    """StreamHandler to split multiline logging messages."""


class MultiLineFileHandler(MultiLineHandlerMixin, logging.FileHandler):
    """FileHandler to split multiline logging messages."""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Benoit Pauwels
"""SciPy linear programming library wrapper."""
from numpy import isfinite
from scipy.optimize import linprog

from gemseo.algos.opt.core.linear_constraints import build_constraints_matrices
from gemseo.algos.opt.opt_lib import OptimizationLibrary
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.algos.opt_result import OptimizationResult
from gemseo.core.function import MDOLinearFunction
from gemseo.utils.py23_compat import PY2


class ScipyLinprog(OptimizationLibrary):
    """SciPy linear programming library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = True

    REDUNDANCY_REMOVAL = "redundancy removal"
    REVISED_SIMPLEX = "REVISED_SIMPLEX"

    OPTIONS_MAP = {
        OptimizationLibrary.MAX_ITER: "maxiter",
        OptimizationLibrary.VERBOSE: "disp",
        REDUNDANCY_REMOVAL: "rr",
    }

    def __init__(self):
        """Constructor.

        Generate the library dictionary that contains the list of algorithms with their
        characteristics.
        """
        super(ScipyLinprog, self).__init__()
        doc = "https://docs.scipy.org/doc/scipy/reference/"

        self.lib_dict = {
            "LINEAR_INTERIOR_POINT": {
                self.INTERNAL_NAME: "interior-point",
                self.DESCRIPTION: "Linear programming by the interior-point"
                " method implemented in the SciPy library",
                self.WEBSITE: doc + "optimize.linprog-interior-point.html",
            },
            ScipyLinprog.REVISED_SIMPLEX: {
                self.INTERNAL_NAME: "revised simplex",
                self.DESCRIPTION: "Linear programming by a two-phase revised"
                " simplex algorithm implemented in the SciPy library",
                self.WEBSITE: doc + "optimize.linprog-revised_simplex.html",
            },
            "SIMPLEX": {
                self.INTERNAL_NAME: "simplex",
                self.DESCRIPTION: "Linear programming by the two-phase simplex"
                " algorithm implemented in the SciPy library",
                self.WEBSITE: doc + "optimize.linprog-simplex.html",
            },
        }
        if PY2:
            # The "revised simplex" algorithm is not available in Python 2.
            # Indeed this feature appeared in SciPy 1.3.0 which requires Python 3.5+.
            # https://docs.scipy.org/doc/scipy-1.3.0/reference/release.1.3.0.html
            del self.lib_dict["REVISED_SIMPLEX"]
        common_items = {
            self.PROBLEM_TYPE: OptimizationProblem.LINEAR_PB,
            self.POSITIVE_CONSTRAINTS: False,
            self.HANDLE_EQ_CONS: True,
            self.HANDLE_INEQ_CONS: True,
        }
        for algo_dict in self.lib_dict.values():
            algo_dict.update(common_items)

    def _get_options(  # pylint: disable=W0221
        self,
        max_iter=999,
        autoscale=False,
        presolve=True,
        redundancy_removal=True,
        callback=None,
        verbose=False,
        normalize_design_space=True,
        **kwargs
    ):
        """Retrieve the options of the library. Defines default values for options using
        keyword arguments.

        :param max_iter: maximum number of iterations, i.e. unique calls to the
            objective function
        :type max_iter: int, optional
        :param autoscale: if True then the linear problem is scaled
            Refer to the SciPy documentation for more details.
        :type autoscale: bool, optional
        :param presolve: if True then attempt to detect infeasibility,
            unboundedness or problem simplifications before solving
            Refer to the SciPy documentation for more details.
        :type presolve: bool, optional
        :param redundancy_removal: if True then linearly dependent
            equality-constraints are removed
        :type redundancy_removal: bool, optional
        :param verbose: if True then convergence messages are printed
        :type verbose: bool, optional
        :param callback: function to be called at least once per iteration
            Takes a scipy.optimize.OptimizeResult as single argument.
            Refer to the SciPy documentation for more details.
        :type callback: callable, optional
        :param kwargs: other algorithms options
        :type kwargs: kwargs
        """
        normalize_ds = normalize_design_space
        options = self._process_options(
            max_iter=max_iter,
            autoscale=autoscale,
            presolve=presolve,
            redundancy_removal=redundancy_removal,
            verbose=verbose,
            callback=callback,
            normalize_design_space=normalize_ds,
            **kwargs
        )
        return options

    def _run(self, **options):
        """Run the algorithm.

        :param options: options dictionary for the algorithm
        :type options: dict
        :returns: the optimization result
        :rtype: OptimizationResult
        """
        # Remove the normalization option from the algorithm options
        options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)

        # Get the starting point and bounds
        x_0, l_b, u_b = self.get_x0_and_bounds_vects(False)
        # Replace infinite bounds with None
        l_b = [val if isfinite(val) else None for val in l_b]
        u_b = [val if isfinite(val) else None for val in u_b]
        bounds = list(zip(l_b, u_b))

        # Build the functions matrices
        # N.B. use the non-processed functions to access the coefficients
        obj_coeff = self.problem.nonproc_objective.coefficients[0, :].real
        constraints = self.problem.nonproc_constraints
        ineq_lhs, ineq_rhs = build_constraints_matrices(
            constraints, MDOLinearFunction.TYPE_INEQ
        )
        eq_lhs, eq_rhs = build_constraints_matrices(
            constraints, MDOLinearFunction.TYPE_EQ
        )

        # |g| is in charge of ensuring max iterations, since it may
        # have a different definition of iterations, such as for SLSQP
        # for instance which counts duplicate calls to x as a new iteration
        options[self.OPTIONS_MAP[self.MAX_ITER]] = 10000000

        # For the "revised simplex" algorithm (available since SciPy 1.3.0 which
        # requires Python 3.5+) the initial guess must be a basic feasible solution,
        # or BFS (geometrically speaking, a vertex of the feasible polyhedron).
        # Here the passed initial guess is always ignored.
        # (A BFS will be automatically looked for during the first phase of the simplex
        # algorithm.)
        linprog_result = linprog(
            c=obj_coeff,
            A_ub=ineq_lhs,
            b_ub=ineq_rhs,
            A_eq=eq_lhs,
            b_eq=eq_rhs,
            bounds=bounds,
            method=self.internal_algo_name,
            options=options,
        )

        # Gather the optimization results
        x_opt = linprog_result.x
        # N.B. SciPy tolerance on bounds is higher than the DesignSpace one
        x_opt = self.problem.design_space.project_into_bounds(x_opt)
        val_opt, jac_opt = self.problem.evaluate_functions(
            x_vect=x_opt,
            eval_jac=True,
            eval_obj=True,
            normalize=False,
            no_db_no_norm=True,
        )
        f_opt = val_opt[self.problem.objective.outvars[0]]
        constraints_values = {
            key: val_opt[key] for key in self.problem.get_constraints_names()
        }
        constraints_grad = {
            key: jac_opt[key] for key in self.problem.get_constraints_names()
        }
        is_feasible = self.problem.is_point_feasible(val_opt)
        optim_result = OptimizationResult(
            x_0=x_0,
            x_opt=x_opt,
            f_opt=f_opt,
            status=linprog_result.status,
            constraints_values=constraints_values,
            constraints_grad=constraints_grad,
            optimizer_name=self.algo_name,
            message=linprog_result.message,
            n_obj_call=None,
            n_grad_call=None,
            n_constr_call=None,
            is_feasible=is_feasible,
        )

        return optim_result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Aerostructure MDO problem's Design Space
****************************************
"""
from __future__ import division, unicode_literals

import numpy as np

from gemseo.algos.design_space import DesignSpace


class AerostructureDesignSpace(DesignSpace):
    """**AerostructureDesignSpace** creates the :class:`.DesignSpace`
    of the Aerostructure problem
    whose :class:`.MDODiscipline` are :class:`.Aerodynamics`,
    :class:`.Structure` and :class:`.Mission`.

    - thick_airfoils belongs to [5., 25.], with initial value equal to 15.
    - thick_panels belongs to [1., 20.], with initial value equal to 3.
    - sweep belongs to [10., 35.], with initial value equal to 25.
    - drag belongs to [100., 1000.], with initial value equal to 340.
    - forces belongs to [-1000., 1000.], with initial value equal to 400.
    - lift belongs to [0.1, 1.0], with initial value equal to 0.5
    - mass belongs to [100000., 500000.], with initial value equal to 100000.
    - reserve_fact belongs to [1000., 1000.], with initial value equal to 0.
    """

    def __init__(self):
        """Constructor."""
        super(AerostructureDesignSpace, self).__init__()

        # construct a dictionary with initial solution
        drag = np.array([340.0], dtype=np.complex128)
        forces = np.array([400.0], dtype=np.complex128)
        lift = np.array([0.5], dtype=np.complex128)
        mass = np.array([100000.0], dtype=np.complex128)
        displ = np.array([-700.0], dtype=np.complex128)
        sweep = np.array([25.0], dtype=np.complex128)
        thick_airfoils = np.array([15.0], dtype=np.complex128)
        thick_panels = np.array([3.0], dtype=np.complex128)
        reserve_fact = np.array([0.0], dtype=np.complex128)

        # design variables
        self.add_variable("thick_airfoils", 1, l_b=5.0, u_b=25.0, value=thick_airfoils)
        self.add_variable("thick_panels", 1, l_b=1.0, u_b=20.0, value=thick_panels)

        # shared design variables
        self.add_variable("sweep", 1, l_b=10.0, u_b=35.0, value=sweep)

        # target coupling variables
        self.add_variable("drag", 1, l_b=100.0, u_b=1000.0, value=drag)
        self.add_variable("forces", 1, l_b=-1000.0, u_b=1000.0, value=forces)
        self.add_variable("lift", 1, l_b=0.1, u_b=1.0, value=lift)
        self.add_variable("mass", 1, l_b=100000.0, u_b=500000.0, value=mass)
        self.add_variable("displ", 1, l_b=-1000.0, u_b=1000.0, value=displ)
        self.add_variable(
            "reserve_fact", 1, l_b=-1000.0, u_b=1000.0, value=reserve_fact
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains the base classes for clustering algorithms.

The :mod:`~gemseo.mlearning.cluster.cluster` module
implements the concept of clustering models,
a kind of unsupervised machine learning algorithm
where the goal is to group data into clusters.
Wherever possible,
these methods should be able to predict the class of the new data,
as well as the probability of belonging to each class.

This concept is implemented
through the :class:`.MLClusteringAlgo` class,
which inherits from the :class:`.MLUnsupervisedAlgo` class,
and through the :class:`.MLPredictiveClusteringAlgo` class
which inherits from :class:`.MLClusteringAlgo`.
"""
from __future__ import division, unicode_literals

from typing import Dict, Iterable, List, NoReturn, Optional, Union

from numpy import atleast_2d, ndarray, unique, zeros

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, MLAlgoParameterType
from gemseo.mlearning.core.ml_algo import SavedObjectType as MLAlgoSavedObjectType
from gemseo.mlearning.core.ml_algo import TransformerType
from gemseo.mlearning.core.unsupervised import MLUnsupervisedAlgo
from gemseo.utils.data_conversion import DataConversion

SavedObjectType = Union[MLAlgoSavedObjectType, ndarray, int]


class MLClusteringAlgo(MLUnsupervisedAlgo):
    """Clustering algorithm.

    The inheriting classes shall overload the
    :meth:`!MLUnsupervisedAlgo._fit` method.

    Attributes:
        labels (List(int)): The indices of the clusters for the different samples.
        n_clusters (int): The number of clusters.
    """

    def __init__(
        self,
        data,  # type:Dataset
        transformer=None,  # type: Optional[TransformerType]
        var_names=None,  # type: Optional[Iterable[str]]
        **parameters  # type:MLAlgoParameterType
    ):  # type: (...) -> None
        super(MLClusteringAlgo, self).__init__(
            data, transformer=transformer, var_names=var_names, **parameters
        )
        self.labels = None
        self.n_clusters = None

    def learn(
        self,
        samples=None,  # type: Optional[List[int]]
    ):  # type: (...) -> None
        super(MLClusteringAlgo, self).learn(samples=samples)
        if self.labels is None:
            raise ValueError("self._fit() shall assign labels.")
        self.n_clusters = unique(self.labels).shape[0]

    def _get_objects_to_save(self):  # type: (...) -> Dict[str,SavedObjectType]
        objects = super(MLClusteringAlgo, self)._get_objects_to_save()
        objects["labels"] = self.labels
        objects["n_clusters"] = self.n_clusters
        return objects


class MLPredictiveClusteringAlgo(MLClusteringAlgo):
    """Predictive clustering algorithm.

    The inheriting classes shall overload the
    :meth:`!MLUnsupervisedAlgo._fit` method, and the
    :meth:`!MLClusteringAlgo._predict` and
    :meth:`!MLClusteringAlgo._predict_proba` methods if possible.
    """

    def predict(
        self,
        data,  # type: DataType
    ):  # type: (...) -> Union[int,ndarray]
        """Predict the clusters from the input data.

        The user can specify these input data either as a NumPy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        If the numpy arrays are of dimension 2,
        their i-th rows represent the input data of the i-th sample;
        while if the numpy arrays are of dimension 1,
        there is a single sample.

        The type of the output data and the dimension of the output arrays
        will be consistent
        with the type of the input data and the dimension of the input arrays.

        Args:
            data: The input data.

        Returns:
            The predicted cluster for each input data sample.
        """
        as_dict = isinstance(data, dict)
        if as_dict:
            data = DataConversion.dict_to_array(data, self.var_names)
        single_sample = len(data.shape) == 1
        data = atleast_2d(data)
        parameters = self.learning_set.DEFAULT_GROUP
        if parameters in self.transformer:
            data = self.transformer[parameters].transform(data)
        clusters = self._predict(atleast_2d(data)).astype(int)
        if single_sample:
            clusters = clusters[0]
        return clusters

    def _predict(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Predict the clusters from input data.

        Args:
            data: The input data with the shape (n_samples, n_inputs).

        Returns:
            The predicted clusters with shape (n_samples,).
        """
        raise NotImplementedError

    def predict_proba(
        self,
        data,  # type: DataType
        hard=True,  # type: bool
    ):  # type: (...)-> ndarray
        """Predict the probability of belonging to each cluster from input data.

        The user can specified these input data either as a numpy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        If the numpy arrays are of dimension 2,
        their i-th rows represent the input data of the i-th sample;
        while if the numpy arrays are of dimension 1,
        there is a single sample.

        The dimension of the output array
        will be consistent with the dimension of the input arrays.

        Args:
            data: The input data.
            hard: Whether clustering should be hard (True) or soft (False).

        Returns:
            The probability of belonging to each cluster,
            with shape (n_samples, n_clusters) or (n_clusters,).
        """
        as_dict = isinstance(data, dict)
        if as_dict:
            data = DataConversion.dict_to_array(data, self.var_names)
        single_sample = len(data.shape) == 1
        data = atleast_2d(data)
        probas = self._predict_proba(atleast_2d(data), hard)
        if single_sample:
            probas = probas.ravel()
        return probas

    def _predict_proba(
        self,
        data,  # type: ndarray
        hard=True,  # type: bool
    ):  # type: (...)-> ndarray
        """Predict the probability of belonging to each cluster.

        Args:
            data: The input data with shape (n_samples, n_inputs).
            hard: Whether clustering should be hard (True) or soft (False).

        Returns:
            The probability of belonging to each cluster
                with shape (n_samples, n_clusters).
        """
        if hard:
            probas = self._predict_proba_hard(data)
        else:
            probas = self._predict_proba_soft(data)
        return probas

    def _predict_proba_hard(
        self,
        data,  # type: ndarray
    ):  # type: (...)-> ndarray
        """Return 1 if the data belongs to a cluster, 0 otherwise.

        Args:
            data: The input data with shape (n_samples, n_inputs).

        Returns:
            The indicator of belonging to each cluster
                with shape (n_samples, n_clusters).
        """
        prediction = self._predict(data)
        probas = zeros((data.shape[0], self.n_clusters))
        for i, pred in enumerate(prediction):
            probas[i, pred] = 1
        return probas

    def _predict_proba_soft(
        self,
        data,  # type: ndarray
    ):  # type: (...)-> NoReturn
        """Predict the probability of belonging to each cluster.

        Args:
            The input data with shape (n_samples, n_inputs).

        Returns:
            The probability of belonging to each cluster
                with shape (n_samples, n_clusters).
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Module with a factory to create an instance of :class:`.SensitivityAnalysis`."""

from __future__ import division, unicode_literals

import logging
from typing import List

from gemseo.algos.design_space import DesignSpace
from gemseo.core.discipline import MDODiscipline
from gemseo.core.factory import Factory
from gemseo.uncertainty.sensitivity.analysis import SensitivityAnalysis

LOGGER = logging.getLogger(__name__)


class SensitivityAnalysisFactory(object):
    """Factory to build instances of :class:`.SensitivityAnalysis`.

    At initialization, this factory scans the following modules
    to search for subclasses of this class:

    - the modules located in "gemseo.uncertainty.sensitivity" and its sub-packages,
    - the modules referenced in the "GEMSEO_PATH",
    - the modules referenced in the "PYTHONPATH" and starting with "gemseo_".

    Then, it can check if a class is present or return the list of available classes.

    Lastly, it can create an instance of a class.

    Examples:
        >>> from numpy import pi
        >>> from gemseo.api import create_discipline, create_parameter_space
        >>> from gemseo.uncertainty.sensitivity.factory import (
        ...     SensitivityAnalysisFactory
        ... )
        >>>
        >>> expressions = {"y": "sin(x1)+7*sin(x2)**2+0.1*x3**4*sin(x1)"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x3", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>>
        >>> factory = SensitivityAnalysisFactory()
        >>> analysis = factory.create(
        ...     "MorrisIndices", discipline, parameter_space, n_replicates=5
        ... )
        >>> indices = analysis.compute_indices()
    """

    def __init__(self):  # type: (...) -> None  # noqa: D107
        self.factory = Factory(SensitivityAnalysis, ("gemseo.uncertainty.sensitivity",))

    def create(
        self,
        sensitivity_analysis,  # type:str
        discipline,  # type: MDODiscipline
        parameter_space,  # type: DesignSpace
        **options
    ):  # type:  (...) -> SensitivityAnalysis
        """Create the sensitivity analysis.

        Args:
            sensitivity_analysis (str): The name of a class
                defining a sensitivity analysis.
            discipline (MDODiscipline): A discipline.
            parameter_space (ParameterSpace): A parameter space.
            **options: The options of the sensitivity analysis.

        Returns:
            SensitivityAnalysis: A sensitivity analysis.
        """
        return self.factory.create(
            sensitivity_analysis,
            discipline=discipline,
            parameter_space=parameter_space,
            **options
        )

    @property
    def available_sensitivity_analyses(self):  # type: (...)-> List[str]
        """The available classes for sensitivity analysis."""
        return self.factory.classes

    def is_available(
        self,
        sensitivity_analysis,  # type: str
    ):  # type: (...) -> bool
        """Check the availability of a SensitivityAnalysis child.

        Args:
            sensitivity_analysis: The name of the sensitivity analysis.

        Returns:
            Availability of a sensitivity analysis.

            True if the type of sensitivity analysis is available.
        """
        return self.factory.is_available(sensitivity_analysis)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Most basic grammar implementation."""

import logging

from gemseo.core.grammars.abstract_grammar import AbstractGrammar
from gemseo.core.grammars.errors import InvalidDataException

LOGGER = logging.getLogger(__name__)


class SimpleGrammar(AbstractGrammar):
    """Store the data named and types."""

    def __init__(self, name):
        """Constructor.

        :param name : grammar name
        """
        super(SimpleGrammar, self).__init__(name)
        # Data names list
        self.data_names = []
        # Data types list in the same order as self.data_names
        self.data_types = []

    def load_data(self, data, raise_exception=True):
        """Loads the data dictionary in the grammar and checks it against self
        properties.

        :param data: the input data
        :param raise_exception: if False, no exception is raised
            when data is invalid (Default value = True)
        """
        self.check(data, raise_exception)
        return data

    def check(self, data, raise_exception=True):
        """Checks local data against self properties.

        :param raise_exception: if False, no exception is raised
            when data is invalid (Default value = True)
        """
        failed = False
        if not isinstance(data, dict):
            failed = True
            LOGGER.error("Grammar data is not a dict, in %s", self.name)
            if raise_exception:
                raise InvalidDataException("Invalid data in " + str(self.name))

        for data_type in self.data_types:
            if data_type is not None and not isinstance(data_type, type):
                msg = "Invalid data type in grammar {} , {} is not a type".format(
                    self.name, data_type
                )
                raise TypeError(msg)

        for data_type, data_name in zip(self.data_types, self.data_names):
            if data_name not in data:
                failed = True
                LOGGER.error("Missing input: %s in %s", data_name, self.name)
            elif not isinstance(data[data_name], data_type):
                failed = True
                LOGGER.error(
                    "Wrong input type for: %s in %s got %s instead of %s",
                    data_name,
                    self.name,
                    type(data[data_name]),
                    data_type,
                )
        if failed and raise_exception:
            raise InvalidDataException("Invalid data in " + str(self.name))

    def initialize_from_base_dict(self, typical_data_dict):
        """Initialize the grammar with types and names from a typical data entry.

        :param typical_data_dict: a data dictionary
        """
        self.data_names = []
        self.data_types = []
        for key, value in typical_data_dict.items():
            self.data_names.append(key)
            self.data_types.append(type(value))

    def get_data_names(self):
        """Returns the list of data names.

        :returns: the data names alphabetically sorted
        """
        return self.data_names

    def is_all_data_names_existing(self, data_names):
        """Checks if data_names are present in grammar.

        :param data_names: the data names list
        :returns: True if all data are in grammar
        """
        for data_name in data_names:
            if not self.is_data_name_existing(data_name):
                return False
        return True

    def _update_field(self, data_name, data_type):
        """Updates self properties with a new property of data_name,data_type Adds it if
        self has no property named data_name Updates self.data_types otherwise.

        :param data_name: the name of the property
        :param data_type: the type of the property
        """
        if data_name in self.data_names:
            indx = self.data_names.index(data_name)
            self.data_names[indx] = data_name
            self.data_types[indx] = data_type
        else:
            self.data_names.append(data_name)
            self.data_types.append(data_type)

    def get_type_of_data_named(self, data_name):
        """Gets the associated type to the data named data_name.

        :param data_name: the name of the property
        :returns: data type associated to data_name
        """
        if not self.is_data_name_existing(data_name):
            raise ValueError("Unknown data named :" + str(data_name))
        indx = self.data_names.index(data_name)
        return self.data_types[indx]

    def update_from(self, input_grammar):
        """Adds properties coming from another grammar.

        :param input_grammar: the grammar to take inputs from
        """
        if not isinstance(input_grammar, AbstractGrammar):
            msg = self._get_update_error_msg(self, input_grammar)
            raise TypeError(msg)

        input_grammar = input_grammar.to_simple_grammar()

        for g_name, g_type in zip(input_grammar.data_names, input_grammar.data_types):
            self._update_field(g_name, g_type)

    def update_from_if_not_in(self, input_grammar, exclude_grammar):
        """Adds properties coming from input_grammar if they are not in exclude_grammar.

        :param input_grammar: the grammar to take inputs from
        :param exclude_grammar: exclusion grammar
        """
        if not isinstance(input_grammar, AbstractGrammar) or not isinstance(
            exclude_grammar, AbstractGrammar
        ):
            msg = self._get_update_error_msg(self, input_grammar, exclude_grammar)
            raise TypeError(msg)

        input_grammar = input_grammar.to_simple_grammar()
        exclude_grammar = exclude_grammar.to_simple_grammar()

        for g_name, g_type in zip(input_grammar.data_names, input_grammar.data_types):
            if exclude_grammar.is_data_name_existing(g_name):
                ex_data_type = exclude_grammar.get_type_of_data_named(g_name)
                if g_type != ex_data_type:
                    raise ValueError(
                        "Inconsistent grammar update "
                        + str(g_type)
                        + " !="
                        + str(ex_data_type)
                    )
            else:
                self._update_field(g_name, g_type)

    def is_data_name_existing(self, data_name):
        """Checks if data_name is present in grammar.

        :param data_name: the data name
        :returns: True if data is in grammar
        """
        return data_name in self.data_names

    def clear(self):
        """Clears the data to produce an empty grammar."""
        self.data_names = []
        self.data_types = []

    def to_simple_grammar(self):
        """Converts to the base SimpleGrammar type.

        :returns: a SimpleGrammar instance equivalent to self
        """
        return self



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Rosenbrock dataset
==================

This :class:`.Dataset` contains 100 evaluations
of the well-known Rosenbrock function:

.. math::

   f(x,y)=(1-x)^2+100(y-x^2)^2

This function is known for its global minimum at point (1,1),
its banana valley and the difficulty to reach its minimum.

This :class:`.Dataset` is based on a full-factorial
design of experiments.

`More information about the Rosenbrock function
<https://en.wikipedia.org/wiki/Rosenbrock_function>`_
"""
from __future__ import division, unicode_literals

from numpy import hstack, linspace, meshgrid

from gemseo.core.dataset import Dataset


class RosenbrockDataset(Dataset):
    """Rosenbrock dataset parametrization."""

    def __init__(
        self,
        name="Rosenbrock",
        by_group=True,
        n_samples=100,
        categorize=True,
        opt_naming=True,
    ):
        """Constructor.

        :param str name: name of the dataset.
        :param bool by_group: if True, store the data by group. Otherwise,
            store them by variables. Default: True
        :param int n_samples: number of samples
        :param bool categorize: distinguish between the different groups of
            variables. Default: True.
        :parma bool opt_naming: use an optimization naming. Default: True.
        """
        super(RosenbrockDataset, self).__init__(name, by_group)
        root_n_samples = int(n_samples ** 0.5)
        x_i = linspace(-2.0, 2.0, root_n_samples)
        x_i, y_i = meshgrid(x_i, x_i)
        x_i = x_i.reshape((-1, 1))
        y_i = y_i.reshape((-1, 1))
        z_i = 100 * (y_i - x_i ** 2) ** 2 + (1 - x_i) ** 2
        data = hstack((x_i, y_i, z_i))
        if categorize:
            if opt_naming:
                groups = {"x": Dataset.DESIGN_GROUP, "rosen": Dataset.FUNCTION_GROUP}
            else:
                groups = {"x": Dataset.INPUT_GROUP, "rosen": Dataset.OUTPUT_GROUP}
        else:
            groups = None
        self.set_from_array(data, ["x", "rosen"], {"x": 2, "rosen": 1}, groups=groups)
        self.set_metadata("root_n_samples", root_n_samples)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Generator of the HTML file containing a D3.js version of the N2 chart."""
from __future__ import unicode_literals

import json
import webbrowser
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from gemseo.core.coupling_structure import DependencyGraph

from gemseo.utils.n2d3.n2_json import N2JSON
from gemseo.utils.py23_compat import Path


class N2HTML(object):
    """Generate a HTML file to visualize a dynamic and interactive N2 chart."""

    def __init__(
        self,
        file_path="n2.html",  # type: Union[str,Path]
        open_browser=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            file_path: The file path of the HTML file.
            open_browser: If True, open the browser and display the HTML file.
        """
        self.__file_path = Path(file_path)
        self.__open_browser = open_browser

    def __create_html_file(
        self,
        json_structure,  # type: str
    ):  # type: (...) -> None
        """Build the HTML file from the JSON structure of the N2 chart.

        Args:
            json_structure: The JSON structure of the N2 chart.
        """
        with Path(self.__file_path).open("w", encoding="utf-8", newline="") as stream:
            stream.write(self.__create_html_contents(json_structure))
        if self.__open_browser:
            webbrowser.open_new_tab(str(self.__file_path))

    def from_graph(
        self,
        graph,  # type: DependencyGraph
    ):  # type: (...) -> None
        """Create the HTML file from a dependency graph.

        Args:
            graph: The dependency graph.
        """
        self.__create_html_file(N2JSON(graph))

    def from_json(
        self,
        file_path,  # type: Union[str,Path]
    ):  # type: (...) -> None
        """Create the HTML file from a JSON file.

        Args:
            file_path: The JSON file containing the JSON structure of the N2 chart.
        """
        with Path(file_path).open(encoding="utf-8") as input_file:
            self.__create_html_file(json.dumps(json.load(input_file)))

    def __create_html_contents(
        self,
        json_data,  # type: str
    ):  # type: (...) -> str
        """Create the HTML content related to the N2 chart.

        Args:
            json_data: The JSON structure of the N2 chart.

        Returns:
            The HTML content.
        """

        css_files = ["style.css", "modal.css", "button.css"]
        js_files = [
            "d3.v3.js",
            "d3.parcoords.js",
            "science.v1.js",
            "tiny-queue.js",
            "reorder.v1.js",
            "matrix.js",
            "expand_groups.js",
            "editable_span.js",
            "canvas_toBlob.js",
            "FileSave.js",
            "save_json.js",
            "save_png.js",
        ]
        template = (Path(__file__).parent / "n2_html.tmpl").read_text()
        data = [
            self.__get_file_contents(Path("css") / css_file) for css_file in css_files
        ]
        data += [self.__get_file_contents(Path("js") / js_file) for js_file in js_files]
        data += [json_data]
        return template.format(*data)

    @staticmethod
    def __get_file_contents(
        file_name,  # type: Path
    ):  # type: (...) -> str
        """Read the content of a file located in the directory `n2d3`.

        Args:
            file_name: The name of the file.

        Returns:
            The content of the file.
        """
        return (Path(__file__).parent / file_name).read_text(encoding="utf-8")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted log-normal distribution."""

from typing import Tuple

from numpy import exp, ndarray

from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceIntervalSide,
)
from gemseo.uncertainty.statistics.tolerance_interval.normal import (
    NormalToleranceInterval,
)


class LogNormalToleranceInterval(NormalToleranceInterval):
    """Computation of tolerance intervals from a data-fitted log-normal distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """

    def __init__(
        self,
        size,  # type: int
        mean,  # type: float
        std,  # type: float
        location,  # type: float
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            mean: The estimation of the mean of the natural logarithm
                of a log-normal distributed random variable.
            std: The estimation of the standard deviation of the natural logarithm
                of a log-normal distributed random variable.
            location: The estimation of the location of the log-normal distributed.
        """
        super(LogNormalToleranceInterval, self).__init__(size, mean, std)
        self.__location = location

    def compute(
        self,
        coverage,  # type: float
        confidence=0.95,  # type: float
        side=ToleranceIntervalSide.BOTH,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Tuple[ndarray,ndarray]
        # noqa: D102
        lower, upper = super(LogNormalToleranceInterval, self).compute(
            coverage, confidence, side
        )
        return exp(lower) + self.__location, exp(upper) + self.__location



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The mean squared error to measure the quality of a regression algorithm.

The :mod:`~gemseo.mlearning.qual_measure.mse_measure` module
implements the concept of mean squared error measures
for machine learning algorithms.

This concept is implemented through the :class:`.MSEMeasure` class
and overloads the :meth:`!MLErrorMeasure._compute_measure` method.

The mean squared error (MSE) is defined by

.. math::

    \\operatorname{MSE}(\\hat{y})=\\frac{1}{n}\\sum_{i=1}^n(\\hat{y}_i-y_i)^2,

where :math:`\\hat{y}` are the predictions and :math:`y` are the data points.
"""
from __future__ import division, unicode_literals

from typing import Union

from numpy import ndarray
from sklearn.metrics import mean_squared_error

from gemseo.mlearning.qual_measure.error_measure import MLErrorMeasure
from gemseo.mlearning.regression.regression import MLRegressionAlgo


class MSEMeasure(MLErrorMeasure):
    """The Mean Squared Error measure for machine learning."""

    def __init__(
        self,
        algo,  # type: MLRegressionAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for regression.
        """
        super(MSEMeasure, self).__init__(algo)

    def _compute_measure(
        self,
        outputs,  # type: ndarray
        predictions,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        multioutput = "raw_values" if multioutput else "uniform_average"
        return mean_squared_error(outputs, predictions, multioutput=multioutput)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Hessian matrix approximations from gradient pairs."""
from __future__ import division, unicode_literals

import logging

from numpy import array, atleast_2d, concatenate, cumsum
from numpy import diag as np_diag
from numpy import dot, eye, inf, sqrt, trace, zeros
from numpy.linalg import LinAlgError, cholesky, inv, multi_dot, norm
from numpy.matlib import repmat
from scipy.optimize import leastsq

LOGGER = logging.getLogger(__name__)


class HessianApproximation(object):
    """Abstract class for hessian approximations from optimization history."""

    def __init__(self, history):
        """Constructor.

        :param history: the optimization history
        """
        self.history = history
        self.x_ref = None
        self.fgrad_ref = None
        self.f_ref = None
        self.b_mat_history = []
        self.h_mat_history = []

    def get_x_grad_history(
        self,
        funcname,
        first_iter=0,
        last_iter=0,
        at_most_niter=-1,
        func_index=None,
        normalize_design_space=False,
        design_space=None,
    ):
        """Get gradient history and design variables history of gradient evaluations.

        :param funcname: function name
        :param first_iter: first iteration after which the history is
            extracted (Default value = 0)
        :param last_iter: last iteration before which the history is
            extracted (Default value = 0)
        :param at_most_niter: maximum number of iterations to take
            (Default value = -1)
        :param func_index: Default value = None)
        :param normalize_design_space: if True, scale the values
            to work in a normalized design space (x between 0 and 1)
        :param design_space: the design space used to scale all values
            mandatory if normalize_design_space==True
        """
        x_grad_hist, x_hist = self.history.get_func_grad_history(funcname, x_hist=True)
        if normalize_design_space:
            (
                x_hist,
                x_grad_hist,
            ) = self._normalize_x_g(x_hist, x_grad_hist, design_space)
        assert len(x_grad_hist) == len(x_hist)
        n_pairs = len(x_grad_hist)
        if n_pairs < 2:
            raise ValueError(
                "Cannot build approximation for function : {}"
                " because its gradient history "
                "is too small : {}.".format(funcname, n_pairs)
            )
        #         LOGGER.info("Building Hessian approximation with " +
        #                     str(n_pairs) + " pairs of x and gradients")
        x_grad_hist = array(x_grad_hist)
        x_hist = array(x_hist)
        n_iter, nparam = x_hist.shape
        grad_shape = x_grad_hist.shape
        n_iter_g = grad_shape[0]
        nparam_g = grad_shape[-1]
        # Function is a vector, jacobian is a 2D matrix
        if len(grad_shape) == 3:
            if func_index is None:
                raise ValueError(
                    "Function {} has a vector output "
                    "then function index of output "
                    "must be specified.".format(funcname)
                )
            if func_index < 0 or func_index >= grad_shape[1]:
                raise ValueError(
                    "Function {} has a vector output of size {},"
                    "function index {} is out of range"
                    ".".format(funcname, grad_shape[1], func_index)
                )
            x_grad_hist = x_grad_hist[:, func_index, :]

        if (n_iter != n_iter_g) or (nparam_g != nparam):
            raise ValueError(
                "Inconsistent gradient and design" " variables optimization history"
            )
        if last_iter == 0:
            x_hist = x_hist[first_iter:, :]
            x_grad_hist = x_grad_hist[first_iter:, :]
        else:
            x_hist = x_hist[first_iter:last_iter, :]
            x_grad_hist = x_grad_hist[first_iter:last_iter, :]

        n_iter = x_hist.shape[0]
        if 0 < at_most_niter < n_iter:
            x_hist = x_hist[n_iter - at_most_niter :, :]
            x_grad_hist = x_grad_hist[n_iter - at_most_niter :, :]
        n_iter, nparam = x_hist.shape
        if n_iter < 2 or nparam == 0:
            raise ValueError(
                "Insufficient optimization history size, "
                "niter={} nparam = {}.".format(n_iter, nparam)
            )
        self.x_ref = x_hist[-1]
        self.fgrad_ref = x_grad_hist[-1]
        if last_iter == 0:
            self.f_ref = array(self.history.get_func_history(funcname))[-1]
        else:
            self.f_ref = array(self.history.get_func_history(funcname))[:last_iter][-1]
        return x_hist, x_grad_hist, n_iter, nparam

    @staticmethod
    def _normalize_x_g(x_hist, x_grad_hist, design_space):
        """scale the values to work in a normalized design space (x between 0 and 1)

        :param design_space: the design space used to scale all values
        """
        if design_space is None:
            raise ValueError(
                "Design space must be provided when using "
                "a normalize_design_space option!"
            )
        unnormalize_vect = design_space.unnormalize_vect

        def normalize_gradient(x_vect):
            """Unnormalize a gradient."""
            return unnormalize_vect(x_vect, minus_lb=False, no_check=True)

        x_scaled, grad_scaled = [], []
        for x_v, g_v in zip(x_hist, x_grad_hist):
            x_scaled.append(design_space.normalize_vect(x_v))
            grad_scaled.append(normalize_gradient(g_v))
        return x_scaled, grad_scaled

    @staticmethod
    def get_s_k_y_k(x_hist, x_grad_hist, iteration):
        """Generate the s_k and y_k terms, respectively design variable difference and
        gradients difference between iterates.

        :param x_hist: design variables history array
        :param x_grad_hist: gradients history array
        :param iteration: iteration number for which the pair must be generated
        """
        n_iter = x_grad_hist.shape[0]
        if iteration >= n_iter:
            raise ValueError(
                "Iteration {} is higher than number of gradients "
                "in the database : {}.".format(iteration, n_iter)
            )
        s_k = atleast_2d(x_hist[iteration + 1] - x_hist[iteration]).T
        y_k = atleast_2d(x_grad_hist[iteration + 1] - x_grad_hist[iteration]).T
        return s_k, y_k

    @staticmethod
    def iterate_s_k_y_k(x_hist, x_grad_hist):
        """Generate the s_k and y_k terms, respectively design variable difference and
        gradients difference between iterates.

        :param x_hist: design variables history array
        :param x_grad_hist: gradients history array
        """
        n_iter = x_hist.shape[0]
        for k in range(n_iter - 1):
            s_k, y_k = HessianApproximation.get_s_k_y_k(x_hist, x_grad_hist, k)
            yield s_k, y_k

    def build_approximation(
        self,
        funcname,
        save_diag=False,
        first_iter=0,
        last_iter=-1,
        b_mat0=None,
        at_most_niter=-1,
        return_x_grad=False,
        func_index=None,
        save_matrix=False,
        scaling=False,
        normalize_design_space=False,
        design_space=None,
    ):  # pylint: disable=W0221
        """Builds the hessian approximation B.

        :param funcname: function name
        :param save_diag: if True, returns the list of diagonal approximations
            (Default value = False)
        :param first_iter: first iteration after which the history is extracted
            (Default value = 0)
        :param last_iter: last iteration before which the history is extracted
            (Default value = -1)
        :param b_mat0: initial approximation matrix (Default value = None)
        :param at_most_niter: maximum number of iterations to take
            (Default value = -1)
        :param return_x_grad: if True, also returns the last gradient and x
            (Default value = False)
        :param func_index: Default value = None)
            (Default value = False)
        :param scaling: do scaling step
        :param normalize_design_space: if True, scale the values
            to work in a normalized design space (x between 0 and 1)
        :param design_space: the design space used to scale all values
            mandatory if normalize_design_space==True
        :returns: the B matrix, its diagonal,
            and eventually the x and grad history pairs
            used to build B, if return_x_grad=True,
            otherwise, None and None are returned for args consistency
        """
        normalize_ds = normalize_design_space
        x_hist, x_grad_hist, _, _ = self.get_x_grad_history(
            funcname,
            first_iter,
            last_iter,
            at_most_niter,
            func_index,
            normalize_ds,
            design_space,
        )
        if b_mat0 is None:
            last_n_grad = x_grad_hist.shape[0] - 2
            s_k, y_k = self.get_s_k_y_k(x_hist, x_grad_hist, last_n_grad)
            alpha = dot(y_k.T, s_k) / dot(y_k.T, y_k)
            b_mat = (1.0 / alpha) * eye(len(x_grad_hist[0]))
        elif b_mat0.size == 0:
            n_x = len(x_hist[0])
            b_mat = zeros((n_x, n_x))
        else:
            b_mat = b_mat0
        diag = []
        for s_k, y_k in self.iterate_s_k_y_k(x_hist, x_grad_hist):
            self.iterate_approximation(b_mat, s_k, y_k, scaling=scaling)
            if save_diag:
                diag.append(np_diag(b_mat).copy())
            if save_matrix:
                self.b_mat_history.append(b_mat.copy())
        if return_x_grad:
            return b_mat, diag, x_hist[-1, :], x_grad_hist[-1, :]
        return b_mat, diag, None, None

    @staticmethod
    def compute_scaling(hessk, hessk_dsk, dskt_hessk_dsk, dyk, dyt_dsk):
        """Compute scaling.

        :param hessk: previous approximation
        :param hessk_dsk: product between hessk and dsk
        :param dskt_hessk_dsk: product between dsk^t, hessk and dsk
        :param dyk: gradients difference between iterates
        :param dyt_dsk: product between dyk^t and dsk^t
        """
        coeff1 = (len(hessk_dsk) - 1) / (
            trace(hessk) - norm(hessk_dsk) ** 2 / dskt_hessk_dsk
        )
        coeff2 = dyt_dsk / norm(dyk) ** 2
        return coeff1, coeff2

    @staticmethod
    def iterate_approximation(hessk, dsk, dyk, scaling=False):
        """BFGS iteration from step k to step k+1.

        :param hessk: previous approximation
        :param dsk: design variable difference between iterates
        :param dyk: gradients difference between iterates
        :param scaling: do scaling step
        :returns: updated approximation
        """
        dyt_dsk = dot(dyk.T, dsk)
        hessk_dsk = dot(hessk, dsk)
        dskt_hessk_dsk = multi_dot((dsk.T, hessk, dsk))
        # Build the next approximation:
        b_first_term = hessk - multi_dot((hessk, dsk, dsk.T, hessk)) / dskt_hessk_dsk
        b_second_term = dot(dyk, dyk.T) / dyt_dsk
        if not scaling:
            hessk[:, :] = b_first_term + b_second_term
        else:
            c_1, c_2 = HessianApproximation.compute_scaling(
                hessk, hessk_dsk, dskt_hessk_dsk, dyk, dyt_dsk
            )
            hessk[:, :] = c_1 * b_first_term + c_2 * b_second_term

    def build_inverse_approximation(
        self,
        funcname,
        save_diag=False,
        first_iter=0,
        last_iter=-1,
        h_mat0=None,
        at_most_niter=-1,
        return_x_grad=False,
        func_index=None,
        save_matrix=False,
        factorize=False,
        scaling=False,
        angle_tol=1e-5,
        step_tol=1e10,
        normalize_design_space=False,
        design_space=None,
    ):
        """Builds the inversed hessian approximation H.

        :param funcname: function name
        :param save_diag: if True, returns the list of diagonal approximations
            (Default value = False)
        :param first_iter: first iteration after which the history is extracted
            (Default value = 0)
        :param last_iter: last iteration before which the history is extracted
            (Default value = -1)
        :param h_mat0: initial inverse approximation matrix
            (Default value = None)
        :param at_most_niter: maximum number of iterations to take
            (Default value = -1)
        :param return_x_grad: if True, also returns the last gradient and x
            (Default value = False)
        :param func_index: Default value = None)
        :param normalize_design_space: if True, scale the values
            to work in a normalized design space (x between 0 and 1)
        :param design_space: the design space used to scale all values
            mandatory if normalize_design_space==True
        :returns: the H matrix, its diagonal,
            and eventually the x and grad history pairs
            used to build H, if return_x_grad=True,
            otherwise, None and None are returned for args consistency
        """
        normalize_ds = normalize_design_space
        x_hist, x_grad_hist, _, _ = self.get_x_grad_history(
            funcname,
            first_iter,
            last_iter,
            at_most_niter,
            func_index,
            normalize_ds,
            design_space,
        )
        h_factor = None  # to become a matrix G such that H = G*G', optionally
        b_factor = None  # to become the inverse of the matrix G
        if h_mat0 is None:
            last_n_grad = x_grad_hist.shape[0] - 2
            s_k, y_k = self.get_s_k_y_k(x_hist, x_grad_hist, last_n_grad)
            alpha = dot(y_k.T, s_k) / dot(y_k.T, y_k)
            n_x = len(x_grad_hist[0])
            h_mat = alpha * eye(n_x)
            b_mat = 1.0 / alpha * eye(n_x)
            if factorize:
                h_factor = sqrt(alpha) * eye(n_x)
                b_factor = eye(n_x) / sqrt(alpha)
        elif len(h_mat0) == 0:
            n_x = len(x_hist[0])
            h_mat = zeros((n_x, n_x))
            b_mat = zeros((n_x, n_x))
            if factorize:
                h_factor = zeros((n_x, n_x))
                b_factor = zeros((n_x, n_x))
        else:
            h_mat = h_mat0
            try:
                b_mat = inv(h_mat)
            except LinAlgError:
                raise LinAlgError("The inversion of h_mat failed")
            if factorize or scaling:
                try:
                    h_factor = cholesky(h_mat)
                    b_factor = cholesky(b_mat).T
                except LinAlgError:
                    raise LinAlgError(
                        "The Cholesky decomposition of h_factor" " or b_factor failed"
                    )
        diag = []
        count = 0
        k = 0
        for s_k, y_k in self.iterate_s_k_y_k(x_hist, x_grad_hist):
            k = k + 1
            if dot(s_k.T, y_k) > angle_tol and norm(y_k, inf) < step_tol:
                count = count + 1
                self.iterate_inverse_approximation(
                    h_mat,
                    s_k,
                    y_k,
                    h_factor,
                    b_mat,
                    b_factor,
                    factorize=factorize,
                    scaling=scaling,
                )
            if save_diag:
                diag.append(np_diag(h_mat).copy())
            if save_matrix:
                self.h_mat_history.append(h_mat.copy())
        if return_x_grad:
            return h_mat, diag, x_hist[-1, :], x_grad_hist[-1, :], None, None, None
        return h_mat, diag, None, None, h_factor, b_mat, b_factor

    @staticmethod
    def compute_corrections(x_hist, x_grad_hist):
        """Computes the corrections from the history."""
        n_iter = x_hist.shape[0]
        x_corr = x_hist[1:n_iter].T - x_hist[: n_iter - 1].T
        grad_corr = x_grad_hist[1:n_iter].T - x_grad_hist[: n_iter - 1].T
        return x_corr, grad_corr

    @staticmethod
    def rebuild_history(x_corr, x_0, grad_corr, g_0):
        """Computes the history from the corrections."""
        # Rebuild the argument history:
        x_hist = repmat(x_0, x_corr.shape[1], 1) + cumsum(x_corr.T, axis=0)
        x_hist = concatenate((atleast_2d(x_0), x_hist), axis=0)
        # Rebuild the gradient history
        x_grad_hist = repmat(g_0, grad_corr.shape[1], 1) + cumsum(grad_corr.T, axis=0)
        x_grad_hist = concatenate((atleast_2d(g_0), x_grad_hist), axis=0)
        return x_hist, x_grad_hist

    @staticmethod
    def iterate_inverse_approximation(
        h_mat,
        s_k,
        y_k,
        h_factor=None,
        b_mat=None,
        b_factor=None,
        factorize=False,
        scaling=False,
    ):
        """Inverse BFGS iteration.

        :param h_mat: previous approximation
        :param s_k: design variable difference between iterates
        :param y_k: gradients difference between iterates
        :returns: updated inverse approximation
        """
        # Compute the two terms of the non-scaled updated matrix:
        yts = dot(y_k.T, s_k)
        proj = eye(len(s_k)) - dot(s_k, y_k.T) / yts
        h_first_term = multi_dot((proj, h_mat, proj.T))
        h_second_term = dot(s_k, s_k.T) / yts
        b_s = dot(b_mat, s_k)
        st_b_s = dot(s_k.T, b_s)
        # Compute the scaling coefficients:
        if scaling:
            coeff1, coeff2 = HessianApproximation.compute_scaling(
                b_mat, b_s, st_b_s, y_k, yts
            )
        else:
            coeff1, coeff2 = 1.0, 1.0
        # Update the inverse approximation H and, optionally, the factor G:
        h_mat[:, :] = h_first_term / coeff1 + h_second_term / coeff2
        if factorize:
            sst_b = dot(s_k, b_s.T)
            left = proj / sqrt(coeff1) + sst_b / sqrt(coeff2 * yts * st_b_s)
            h_factor[:, :] = dot(left, h_factor)
            # b_factor[:, :] = dot(eye(len(s_k)) - sstB.T / stBs / sqrt(coeff1)
            #                      + dot(y_k, s_k.T)
            #                      / sqrt(coeff2 * stBs * yts),
            #                      b_factor)
            right = sqrt(coeff1) * (eye(len(s_k)) - sst_b / st_b_s)
            right += sqrt(coeff2) * dot(s_k, y_k.T) / sqrt(st_b_s * yts)
            b_factor[:, :] = dot(b_factor, right)

        # Update the Hessian approximation:
        b_first_term = b_mat - multi_dot((b_s, b_s.T)) / st_b_s
        b_second_term = dot(y_k, y_k.T) / yts
        b_mat[:, :] = coeff1 * b_first_term + coeff2 * b_second_term


#             b_mat[:, :] = multi_dot((proj.T, b_mat, proj)) \
#                 + dot(y_k, y_k.T) / yts


class BFGSApprox(HessianApproximation):

    """Builds a BFGS approximation from optimization history."""

    @staticmethod
    def iterate_s_k_y_k(x_hist, x_grad_hist):
        """Generate the s_k and y_k terms, respectively design variable difference and
        gradients difference between iterates.

        :param x_hist: design variables history array
        :param x_grad_hist: gradients history array
        """
        n_iter = x_hist.shape[0]
        for k in range(n_iter - 1):
            s_k, y_k = BFGSApprox.get_s_k_y_k(x_hist, x_grad_hist, k)
            # All pairs curvatures shall be positive
            # if dot(s_k.T, y_k) > 0.:
            if dot(s_k.T, y_k) > 1e-16 * dot(y_k.T, y_k):
                yield s_k, y_k


class SR1Approx(HessianApproximation):

    """Builds a Symmetric Rank One approximation from optimization history."""

    EPSILON = 1e-8

    @staticmethod
    def iterate_approximation(b_mat, s_k, y_k, scaling=False):
        """SR1 iteration.

        :param b_mat: previous approximation
        :param s_k: design variable difference between iterates
        :param y_k: gradients difference between iterates
        :param scaling: do scaling step
        :returns: updated approximation
        """
        d_mat = y_k - multi_dot((b_mat, s_k))
        den = multi_dot((d_mat.T, s_k))
        if abs(den) > SR1Approx.EPSILON * norm(s_k) * norm(d_mat):
            b_mat[:, :] = b_mat + multi_dot((d_mat, d_mat.T)) / den
        else:
            LOGGER.debug(
                "Denominator of SR1 update is too small, " "update s_kipped %s.", den
            )


class LSTSQApprox(HessianApproximation):

    """Builds a Least squares approximation from optimization history."""

    def build_approximation(
        self,
        funcname,
        save_diag=False,
        first_iter=0,
        last_iter=-1,
        b_mat0=None,
        at_most_niter=-1,
        return_x_grad=False,
        scaling=False,
        func_index=-1,
        normalize_design_space=False,
        design_space=None,
    ):
        """Builds the hessian approximation.

        :param funcname: function name
        :param save_diag: if True, returns the list of diagonal approximations
            (Default value = False)
        :param first_iter: first iteration after which the history is extracted
            (Default value = 0)
        :param last_iter: last iteration before which the history is extracted
            (Default value = -1)
        :param b_mat0: initial approximation matrix (Default value = None)
        :param at_most_niter: Default value = -1)
        :param return_x_grad: Default value = False)
        :param func_index: Default value = -1)
        :param normalize_design_space: if True, scale the values
            to work in a normalized design space (x between 0 and 1)
        :param design_space: the design space used to scale all values
            mandatory if normalize_design_space==True
        :returns: the B matrix, its diagonal,
            and eventually the x and grad history pairs
            used to build B, if return_x_grad=True,
            otherwise, None and None are returned for args consistency
        """
        x_hist, x_grad_hist, n_iter, nparam = self.get_x_grad_history(
            funcname,
            first_iter,
            last_iter,
            at_most_niter,
            func_index=func_index,
            normalize_design_space=normalize_design_space,
            design_space=design_space,
        )
        sec_dim = max(nparam, n_iter)
        diag = []
        assert len(x_grad_hist) == len(x_hist)

        def y_to_b(y_vars):
            """Reshapes the approximation from vector to matrix.

            :param y_vars: the vector approximation
            :param y_vars: returns: the matrix shaped approximation
            :returns: the matrix shaped approximation
            """
            y_mat = y_vars.reshape((nparam, nparam))
            return y_mat + y_mat.T

        def func(y_vars):
            """Create the least square function.

            :param y_vars: the current approximation vector
            :param y_vars: returns: the estimated error vector
            :returns: the estimated error vector
            """
            b_mat_current = y_to_b(y_vars)
            err = zeros((nparam, sec_dim))
            for i in range(n_iter):
                x_l = x_hist[i] - self.x_ref
                err[:, i] = dot(b_mat_current, x_l) - x_grad_hist[i]
            err = err.reshape(nparam * sec_dim)
            if n_iter < nparam:  #
                err += y_vars
            return err

        y_0 = zeros(nparam * nparam)
        LOGGER.debug("Start least squares problem..")
        y_opt, ier = leastsq(func, x0=y_0)  # , cov_x, infodict, mesg, ier
        LOGGER.debug("End least squares, msg=%s", str(ier))
        b_mat = y_to_b(y_opt)
        if return_x_grad:
            return b_mat, diag, x_hist[-1, :], x_grad_hist[-1, :]
        return b_mat, diag, None, None



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Abstract factory to create drivers
**********************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.factory import Factory

LOGGER = logging.getLogger(__name__)


class DriverFactory(object):
    """Base class for definition of optimization and/or DOE factory.

    Automates the creation of library interfaces given a name of the algorithm.
    """

    def __init__(self, driver_lib_class, driver_package):
        """Initializes the factory: scans the directories to search for subclasses of
        DriverLib.

        Searches in "GEMSEO_PATH" and gemseo.mda
        """
        self.factory = Factory(driver_lib_class, (driver_package,))
        self.__algo_name_to_lib_name = {}
        self.__update_libdict()

    def __update_libdict(self):
        """Updates the self.__algo_name_to_lib_name dict with available libraries
        list."""
        for lib_name in self.libraries:
            lib = self.create(lib_name)
            for algo_name in lib.algorithms:
                self.__algo_name_to_lib_name[algo_name] = lib_name

    def is_available(self, name):
        """Checks the availability of a library name or algorithm name.

        :param name: the name of the library name or algorithm name
        :returns: True if the library is installed
        """
        is_lib = self.factory.is_available(name)
        return name in self.__algo_name_to_lib_name or is_lib

    @property
    def algorithms(self):
        """Lists the available algorithms names in the present configuration.

        :returns: the list of algorithms as a string list
        """
        return list(self.__algo_name_to_lib_name.keys())

    @property
    def libraries(self):
        """Lists the available library names in the present configuration.

        :returns: the list of libraries as a string list
        """
        return self.factory.classes

    def create(self, name):
        """Factory method to create a DriverLib subclass from algo identifier or a
        library identifier.

        :param name: library or algorithm name
        :type name: string
        :returns: library according to context (optimization or DOE)
        """
        lib_name = self.__algo_name_to_lib_name.get(name)
        algo_name = None
        if lib_name is not None:
            algo_name = name
        elif self.factory.is_available(name):
            lib_name = name
        else:
            algorithms = sorted(self.algorithms)
            raise ImportError(
                "No algorithm or library of algorithms named "
                + str(name)
                + " is available among available "
                + "ones : "
                + str(algorithms)
            )
        lib_created = self.factory.create(lib_name)
        # Set the algo name if it was passed by the user as "name" arg
        lib_created.algo_name = algo_name
        return lib_created

    def execute(self, problem, algo_name, **options):
        """Finds the appropriate library and executes the driver on the problem.

        :param problem: the problem on which to run the execution
        :param algo_name: the algorithm name
        :param options: the options dict for the DOE,
            see associated JSON file
        """
        lib = self.create(algo_name)
        return lib.execute(problem, algo_name=algo_name, **options)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Caching module to avoid multiple evaluations of a discipline
************************************************************
"""
from __future__ import division, unicode_literals

import logging
from os.path import exists

import h5py
from future.utils import with_metaclass
from numpy import append, array, bytes_, unicode_

from gemseo.core.cache import AbstractCache, AbstractFullCache, hash_data_dict, to_real
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.locks import synchronized
from gemseo.utils.multi_processing import RLock
from gemseo.utils.py23_compat import PY2, long, string_array, string_dtype
from gemseo.utils.singleton import SingleInstancePerFileAttribute

LOGGER = logging.getLogger(__name__)


class HDF5Cache(AbstractFullCache):
    """Cache using disk HDF5 file to store the data."""

    def __init__(self, hdf_file_path, hdf_node_path, tolerance=0.0, name=None):
        """Initialize a singleton to access a HDF file. This singleton is used for
        multithreaded/multiprocessing access with a Lock.

        Initialize cache tolerance.
        By default, don't use approximate cache.
        It is up to the user to choose to optimize CPU time with this or not
        could be something like 2 * finfo(float).eps

        Parameters
        ----------
        hdf_file_path : str
            Path of the HDF file.
        hdf_node_path : str
            Node of the HDF file.
        tolerance : float
            Tolerance that defines if two input vectors
            are equal and cached data shall be returned.
            If 0, no approximation is made. Default: 0.
        name : str
            Name of the cache.

        Examples
        --------
        >>> from gemseo.caches.hdf5_cache import HDF5Cache
        >>> cache = HDF5Cache('my_cache.hdf5', 'my_node')
        """
        if not name:
            name = hdf_node_path
        self.__file_path = hdf_file_path
        self._hdf_file = HDF5FileSingleton(hdf_file_path)
        self.__hdf_node_path = hdf_node_path
        super(HDF5Cache, self).__init__(tolerance, name)
        self._read_hashes()

    def __str__(self):
        msg = super(HDF5Cache, self).__str__()
        msg += "\n" + "HDF file path " + str(self.__file_path) + "\n"
        msg += "HDF node path " + str(self.__hdf_node_path)
        return msg

    def _duplicate_from_scratch(self):
        return HDF5Cache(
            "inc_" + self.__file_path, self.__hdf_node_path, self.tolerance, self.name
        )

    def _set_lock(self):
        """Set a lock for multithreading, either from an external object or internally
        by using RLock()."""
        return self._hdf_file.lock

    @synchronized
    def _read_hashes(self):
        """Read the hashes dict in the HDF file."""
        max_grp = self._hdf_file.read_hashes(self._hashes, self.__hdf_node_path)
        self._last_accessed_group.value = max_grp
        self._max_group.value = max_grp
        n_hash = len(self._hashes)
        if n_hash > 0:
            msg = "Found %s entries in the cache file : %s node : %s"
            LOGGER.info(msg, n_hash, self.__file_path, self.__hdf_node_path)

    def _has_group(self, sample_id, var_group):
        """Check if the dataset has the particular variables group filled in.

        :param int sample_id: sample ID.
        :param str var_group: name of the variables group.
        :return: True if the variables group is filled in.
        :rtype: bool
        """
        return self._hdf_file.has_group(sample_id, var_group, self.__hdf_node_path)

    @synchronized
    def clear(self):
        """Clear the cache.

        Examples
        --------
        >>> from gemseo.caches.hdf5_cache import HDF5Cache
        >>> from numpy import array
        >>> cache = HDF5Cache('my_cache.hdf5', 'my_node')
        >>> for index in range(5):
        >>>     data = {'x': array([1.])*index, 'y': array([.2])*index}
        >>>     cache.cache_outputs(data, ['x'], data, ['y'])
        >>> cache.get_length()
        5
        >>> cache.clear()
        >>> cache.get_length()
        0
        """
        super(HDF5Cache, self).clear()
        self._hdf_file.clear(self.__hdf_node_path)

    def _read_data(self, group_number, group_name, h5_open_file=None):
        """Read data from an HDF file.

        :param str group_name: name of the group where data is written.
        :param int group_number: number of the group.
        :param h5_open_file: eventually the already opened file.
            this improves performance but is incompatible with
            multiprocess/treading
        :returns: data dict and jacobian
        """
        result = self._hdf_file.read_data(
            group_number, group_name, self.__hdf_node_path, h5_open_file=h5_open_file
        )[0]
        if group_name == self.JACOBIAN_GROUP and result is not None:
            result = DataConversion.dict_to_jac_dict(result)
        return result

    def _write_data(self, values, names, var_group, sample_id):
        """Write data associated with a variables group and a sample ID into the
        dataset.

        :param dict values: data dictionary where keys are variables names
            and values are variables values (numpy arrays).
        :param list(str) names: list of input data names to write.
        :param str var_group: name of the variables group,
            either AbstractCache.INPUTS_GROUP, AbstractCache.OUTPUTS_GROUP or
            AbstractCache.JACOBIAN_GROUP.
        :param int sample_id: sample ID.
        """
        self._hdf_file.write_data(
            values,
            names,
            var_group,
            group_num=sample_id,
            hdf_node_path=self.__hdf_node_path,
        )

    @synchronized
    def get_data(self, index, **options):
        """Gets the data associated to a sample ID.

        :param str index: sample ID.
        :param options: options passed to the _read_data() method.
        :return: input data, output data and jacobian.
        :rtype: dict
        """
        with h5py.File(self._hdf_file.hdf_file_path, "a") as h5file:
            datum = super(HDF5Cache, self).get_data(index, h5_open_file=h5file)
        return datum

    @synchronized
    def _get_all_data(self):
        """Same as _all_data() but first open a file, then pass the opened file to the
        generator and lastly, close the file."""
        with h5py.File(self._hdf_file.hdf_file_path, "a") as h5file:
            for data in self._all_data(h5_open_file=h5file):
                yield data


class HDF5FileSingleton(with_metaclass(SingleInstancePerFileAttribute, object)):
    """Singleton to access a HDF file Used for multithreaded/multiprocessing access with
    a Lock."""

    # We create a single instance of cache per HDF5 file
    # to allow the multiprocessing lock to work
    # Ie we want a single lock even if we instantiate multiple
    # caches
    HASH_TAG = "hash"
    INPUTS_GROUP = AbstractCache.INPUTS_GROUP
    OUTPUTS_GROUP = AbstractCache.OUTPUTS_GROUP
    JACOBIAN_GROUP = AbstractCache.JACOBIAN_GROUP

    def __init__(self, hdf_file_path):
        """Constructor.

        :param hdf_file_path: path to the HDF5 file
        """
        self.hdf_file_path = hdf_file_path
        # Attach the lock to the file and NOT the Cache
        # because it is a singletton
        self.lock = RLock()

    def write_data(
        self, data, data_names, group_name, group_num, hdf_node_path, h5_open_file=None
    ):
        """Cache input data to avoid re evaluation.

        :param data: the data to cache
        :param data_names: list of data names
        :param group_name: inputs or outputs or jacobian group
        :param hdf_node_path: name of the main HDF group
        :param h5_open_file: eventually the already opened file.
            this improves performance but is incompatible with
            multiprocess/treading
        """
        if h5_open_file is None:
            h5file = h5py.File(self.hdf_file_path, "a")
        else:
            h5file = h5_open_file

        node_group = h5file.require_group(hdf_node_path)
        num_group = node_group.require_group(str(group_num))
        io_group = num_group.require_group(group_name)

        try:
            # Write hash if needed
            if num_group.get(self.HASH_TAG) is None:
                data_hash = string_array([hash_data_dict(data, data_names)])
                num_group.create_dataset(self.HASH_TAG, data=data_hash)

            for data_name in data_names:
                val = data.get(data_name)
                if val is not None:
                    if val.dtype.type is unicode_:
                        io_group.create_dataset(
                            data_name, data=val.astype(string_dtype)
                        )
                    else:
                        io_group.create_dataset(data_name, data=to_real(val))

        # IOError and RuntimeError are for python 2.7
        except (RuntimeError, IOError, ValueError):
            h5file.close()
            raise RuntimeError(
                "Failed to cache dataset %s.%s.%s in file: %s",
                hdf_node_path,
                group_num,
                group_name,
                self.hdf_file_path,
            )

        if h5_open_file is None:
            h5file.close()

    def read_data(self, group_number, group_name, hdf_node_path, h5_open_file=None):
        """Read a data dict in the hdf.

        :param group_name: name of the group where data is written
        :param group_number: number of the group
         :param hdf_node_path: name of the main HDF group
        :returns: data dict and jacobian
        :param h5_open_file: eventually the already opened file.
            this improves performance but is incompatible with
            multiprocess/treading
        """
        if h5_open_file is None:
            h5file = h5py.File(self.hdf_file_path, "r")
        else:
            h5file = h5_open_file

        node = h5file[hdf_node_path]
        if not self._has_group(group_number, group_name, hdf_node_path, h5file):
            return None, None

        number_dataset = node[str(group_number)]
        values_group = number_dataset[group_name]
        data = {key: array(val) for key, val in values_group.items()}
        if group_name == self.INPUTS_GROUP:
            read_hash = number_dataset[self.HASH_TAG]
            data_hash = long(array(read_hash)[0])
        else:
            data_hash = None

        if h5_open_file is None:
            h5file.close()
        ##########################################
        # Python  key.encode("ascii")
        ##########################################
        if PY2:
            data = {key.encode("ascii"): val for key, val in data.items()}

        for key, val in data.items():
            if val.dtype.type is bytes_:
                data[key] = val.astype(unicode_)

        return data, data_hash

    @staticmethod
    def _has_group(group_number, group_name, hdf_node_path, h5file):
        """Check if a group is present in the HDF file.

        :param group_name: name of the group where data is written
        :param group_number: number of the group
        :param hdf_node_path: name of the main HDF group
        :param h5file: the opened HDF File pointer
        :returns: True if the group exists
        """
        node = h5file[hdf_node_path]
        number_dataset = node.get(str(group_number))
        if number_dataset is None:
            return False
        values_group = number_dataset.get(group_name)
        if values_group is None:
            return False
        return True

    def has_group(self, group_number, group_name, hdf_node_path):
        """Check if a group is present in the HDF file.

        :param group_name: name of the group where data is written
        :param group_number: number of the group
        :param hdf_node_path: name of the main HDF group
        :returns: True if the group exists
        """
        with h5py.File(self.hdf_file_path, "r") as h5file:
            has_grp = self._has_group(group_number, group_name, hdf_node_path, h5file)
        return has_grp

    def read_hashes(self, hashes_dict, hdf_node_path):
        """Read the hashes in the HDF file.

        :param hashes_dict: dict of hashes to fill
        :param hdf_node_path: name of the main HDF group
        :return:  max_group
        """
        if not exists(self.hdf_file_path):
            return 0

        # We must lock so that no data is added to the cache meanwhile
        with h5py.File(self.hdf_file_path, "r") as h5file:
            node_group = h5file.get(hdf_node_path)

            if node_group is None:
                return 0

            max_group = 0

            for group_num, group in node_group.items():
                group_num = int(group_num)
                read_hash = group[self.HASH_TAG]
                hash_value = long(array(read_hash)[0])
                get_hash = hashes_dict.get(hash_value)

                if get_hash is None:
                    hashes_dict[hash_value] = array([group_num])
                else:
                    hashes_dict[hash_value] = append(get_hash, array([group_num]))

                max_group = max(max_group, group_num)

        return max_group

    def clear(self, hdf_node_path):
        """Clear the data in the cache.

        :param hdf_node_path: node path to clear
        """
        with h5py.File(self.hdf_file_path, "a") as h5file:
            del h5file[hdf_node_path]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A parallel coordinates plot of functions and x."""

from __future__ import division, unicode_literals

import logging
from typing import Sequence

import matplotlib as mpl
from matplotlib import pyplot
from matplotlib.figure import Figure
from numpy import array, ndarray

from gemseo.post.core.colormaps import PARULA
from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class ParallelCoordinates(OptPostProcessor):
    """Parallel coordinates among design variables, outputs functions and constraints.

    x- and y- figure sizes can be changed in option.
    """

    @staticmethod
    def parallel_coordinates(
        y_data,  # type: ndarray
        x_names,  # type: Sequence[str]
        color_criteria,  # type: Sequence[float]
        figsize_x,  # type: int
        figsize_y,  # type: int
    ):  # type: (...) -> Figure
        """Plot the parallel coordinates.

        Args:
            y_data: The lines data to plot.
            x_names: The names of the abscissa.
            color_criteria: The values of same length as `y_data`
                to colorize the lines.
            figsize_x: The size of the figure in horizontal direction (inches).
            figsize_y: The size of the figure in vertical direction (inches).
        """
        n_x, n_cols = y_data.shape
        assert n_cols == len(x_names)
        assert n_x == len(color_criteria)
        x_values = list(range(n_cols))
        fig = pyplot.figure(figsize=(figsize_x, figsize_y))
        main_ax = pyplot.gca()
        c_max = color_criteria.max()
        c_min = color_criteria.min()
        color_criteria_n = (color_criteria - c_min) / (c_max - c_min)
        for i in range(n_x):
            color = array(PARULA(color_criteria_n[i])).flatten()
            main_ax.plot(x_values, y_data[i, :], c=color)

        norm = mpl.colors.Normalize(vmin=c_min, vmax=c_max)
        cax, _ = mpl.colorbar.make_axes(main_ax)
        mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=PARULA)
        for i in x_values:
            main_ax.axvline(i, linewidth=1, color="black")

        main_ax.set_xticks(x_values)
        main_ax.set_xticklabels(x_names)
        main_ax.grid()
        return fig

    def _plot(
        self,
        figsize_x=10,  # type: int
        figsize_y=2,  # type: int
    ):  # type: (...) -> None
        """
        Args:
            figsize_x: The size of the figure in horizontal direction (inches).
            figsize_y: The size of the figure in vertical direction (inches).
        """
        all_funcs = self.opt_problem.get_all_functions_names()
        design_variables = self.opt_problem.get_design_variable_names()
        vals, vname, _ = self.database.get_history_array(all_funcs, add_dv=True)
        n_x = len(self.database.get_x_by_iter(0))
        x_names = []
        for d_v in design_variables:
            dv_size = self.opt_problem.design_space.variables_sizes[d_v]
            if dv_size == 1:
                x_names.append(d_v)
            else:
                for i in range(dv_size):
                    x_names.append(d_v + "_" + str(i))
        func_names = vname[: len(vname) - n_x]
        obj_name = self.opt_problem.get_objective_name()
        obj_val = vals[:, vname.index(obj_name)]
        x_vals = vals[:, vals.shape[1] - len(x_names) :]
        x_vals = self.opt_problem.design_space.normalize_vect(x_vals)
        fig = self.parallel_coordinates(x_vals, x_names, obj_val, figsize_x, figsize_y)
        fig.suptitle(
            "Design variables history colored" + " by '" + obj_name + "'  value"
        )

        self._add_figure(fig, "para_coord_des_vars")

        func_vals = vals[:, : vals.shape[1] - len(x_names)]
        fig = self.parallel_coordinates(
            func_vals, func_names, obj_val, figsize_x, figsize_y
        )
        fig.suptitle(
            "Objective function and constraints history"
            + " colored by '"
            + obj_name
            + "' value"
        )

        self._add_figure(fig, "para_coord_funcs")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The base class for all formulations."""

from __future__ import division, unicode_literals

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from gemseo.algos.design_space import DesignSpace
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequence
from gemseo.core.json_grammar import JSONGrammar

if TYPE_CHECKING:
    from gemseo.core.scenario import Scenario

import six
from custom_inherit import DocInheritMeta
from numpy import array, copy, empty, in1d, ndarray, where, zeros
from six import string_types

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction, MDOFunctionGenerator

LOGGER = logging.getLogger(__name__)


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class MDOFormulation(object):
    """Abstract MDO formulation class to be extended in subclasses for use.

    This class creates the objective function and constraints from the disciplines.
    It defines the process implicitly.

    By default,

    - the objective function is minimized,
    - the type of a constraint is equality,
    - the activation value of a constraint is 0.

    The link between the instances of :class:`.MDODiscipline`,
    the name of the objective function
    and the names of the constraints
    is made with the :class:`.MDOFunctionGenerator`,
    which generates instances of :class:`.MDOFunction` from the disciplines.
    """

    NAME = "MDOFormulation"

    def __init__(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
        objective_name,  # type: str
        design_space,  # type: DesignSpace
        maximize_objective=False,  # type: bool
        **options
    ):  # type: (...) -> None # pylint: disable=W0613
        """
        Args:
            disciplines: The disciplines.
            objective_name: The name of the objective function.
            design_space: The design space.
            maximize_objective: If True, the objective function is maximized.
            **options: The options of the formulation.
        """
        self.check_disciplines(disciplines)
        self.disciplines = disciplines
        self._objective_name = objective_name
        self.opt_problem = OptimizationProblem(design_space)
        self._maximize_objective = maximize_objective

    @property
    def design_space(self):  # type: (...) -> DesignSpace
        """The design space on which the formulation is applied."""
        return self.opt_problem.design_space

    @staticmethod
    def check_disciplines(
        disciplines,  # type: Any
    ):  # type: (...) -> None
        """Check that the disciplines are provided as a list.

        Args:
            disciplines: The disciplines.
        """
        if not disciplines or not isinstance(disciplines, list):
            raise TypeError("Disciplines must be provided to the formulation as a list")

    @staticmethod
    def _check_add_cstr_input(
        output_name,  # type: str,
        constraint_type,  # type:str
    ):  # type: (...) -> List[str]
        """Check the output name and constraint type passed to :meth:`.add_constraint`.

        Args:
            output_name: The name of the output to be used as a constraint.
                For instance, if g_1 is given and constraint_type="eq",
                g_1=0 will be added as a constraint to the optimizer.
            constraint_type: The type of constraint,
                either "eq" for equality constraint
                or "ineq" for inequality constraint.
        """
        if constraint_type not in [MDOFunction.TYPE_EQ, MDOFunction.TYPE_INEQ]:
            raise ValueError(
                "Constraint type must be either 'eq' or 'ineq',"
                " got: %s instead" % constraint_type
            )
        if isinstance(output_name, list):
            outputs_list = output_name
        else:
            outputs_list = [output_name]
        return outputs_list

    def add_constraint(
        self,
        output_name,  # type: str
        constraint_type=MDOFunction.TYPE_EQ,  # type: str
        constraint_name=None,  # type: Optional[str]
        value=None,  # type: Optional[float]
        positive=False,  # type: bool
    ):  # type: (...) -> None
        """Add a user constraint.

        A user constraint is a design constraint
        in addition to the formulation specific constraints
        such as the targets (a.k.a. consistency constraints) in IDF.

        The strategy of repartition of constraints is defined in the formulation class.

        Args:
            output_name: The name of the output to be used as a constraint.
                For instance, if g_1 is given and constraint_type="eq",
                g_1=0 will be added as a constraint to the optimizer.
            constraint_type: The type of constraint,
                either "eq" for equality constraint or "ineq" for inequality constraint.
            constraint_name: The name of the constraint to be stored,
                If None, the name is generated from the output name.
            value: The value of activation of the constraint.
                If None, the value is equal to 0.
            positive: If True, the inequality constraint is positive.
        """
        outputs_list = self._check_add_cstr_input(output_name, constraint_type)

        mapped_cstr = self._get_function_from(outputs_list, top_level_disc=True)
        mapped_cstr.f_type = constraint_type

        if constraint_name is not None:
            mapped_cstr.name = constraint_name
        self.opt_problem.add_constraint(mapped_cstr, value=value, positive=positive)

    def add_observable(
        self,
        output_names,  # type: Union[str,Sequence[str]]
        observable_name=None,  # type: Optional[str]
        discipline=None,  # type: Optional[MDODiscipline]
    ):  # type: (...) -> None
        """Add an observable to the optimization problem.

        The repartition strategy of the observable is defined in the formulation class.

        Args:
            output_names: The name(s) of the output(s) to observe.
            observable_name: The name of the observable.
            discipline: The discipline computing the observed outputs.
                If None, the discipline is detected from inner disciplines.
        """
        if isinstance(output_names, string_types):
            output_names = [output_names]
        obs_fun = self._get_function_from(
            output_names, top_level_disc=True, discipline=discipline
        )
        if observable_name is not None:
            obs_fun.name = observable_name
        obs_fun.f_type = MDOFunction.TYPE_OBS
        self.opt_problem.add_observable(obs_fun)

    def get_top_level_disc(self):  # type: (...) -> List[MDODiscipline]
        """Return the disciplines which inputs are required to run the scenario.

        A formulation seeks to
        evaluate objective function and constraints from inputs.
        It structures the optimization problem into multiple levels of disciplines.
        The disciplines directly depending on these inputs
        are called top level disciplines.

        By default, this method returns all disciplines.
        This method can be overloaded by subclasses.

        Returns:
            The top level disciplines.
        """
        return self.disciplines

    @staticmethod
    def _get_mask_from_datanames(
        all_data_names,  # type: ndarray
        masked_data_names,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Get a mask of all_data_names for masked_data_names.

        This mask is an array of the size of all_data_names
        with True values when masked_data_names are in all_data_names.

        Args:
            all_data_names: The main array for mask.
            masked_data_names: The array which masks all_data_names.

        Returns:
            A True / False valued mask array.
        """
        places = in1d(all_data_names, masked_data_names)
        return where(places)

    def _get_generator_from(
        self,
        output_names,  # type: Iterable[str]
        top_level_disc=False,  # type: bool
    ):  # type: (...) -> MDOFunctionGenerator
        """Create a generator of :class:`.MDOFunction` from the names of the outputs.

        Find a discipline which computes all the provided outputs
        and build the associated MDOFunctionGenerator.

        Args:
            output_names: The names of the outputs.
            top_level_disc: If True, search outputs among top level disciplines.

        Returns:
            A generator of :class:`.MDOFunction` instances.
        """
        if top_level_disc:
            search_among = self.get_top_level_disc()
        else:
            search_among = self.disciplines
        for discipline in search_among:
            if discipline.is_all_outputs_existing(output_names):
                return MDOFunctionGenerator(discipline)
        raise ValueError(
            "No discipline known by formulation %s"
            " has all outputs named %s" % (type(self).__name__, output_names)
        )

    def _get_generator_with_inputs(
        self,
        input_names,  # type: Iterable[str]
        top_level_disc=False,  # type: bool
    ):  # type: (...) -> MDOFunctionGenerator
        """Create a generator of :class:`.MDOFunction` from the names of the inputs.

        Find a discipline which has all the provided inputs
        and build the associated MDOFunctionGenerator.

        Args:
            input_names: The names of the inputs.
            top_level_disc: If True, search inputs among the top level disciplines.

        Returns:
            A generator of :class:`.MDOFunction` instances.
        """
        if top_level_disc:
            search_among = self.get_top_level_disc()
        else:
            search_among = self.disciplines
        for discipline in search_among:
            if discipline.is_all_inputs_existing(input_names):
                return MDOFunctionGenerator(discipline)
        raise ValueError(
            "No discipline known by formulation %s"
            " has all inputs named %s" % (type(self).__name__, input_names)
        )

    def mask_x(
        self,
        masking_data_names,  # type: Iterable[str]
        x_vect,  # type: ndarray
        all_data_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> ndarray
        """Mask a vector from a subset of names, with respect to a set of names.

        Args:
            masking_data_names: The names of data to keep.
            x_vect: The vector of float to mask.
            all_data_names: The set of all names.
                If None, use the design variables stored in the design space.

        Returns:
            A boolean mask with the same shape as the input vector.
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        i_min = 0
        x_mask = array([False] * x_vect.size)
        for key in all_data_names:
            var_length = self._get_dv_length(key)
            i_max = i_min + var_length
            if len(x_vect) < i_max:
                raise ValueError(
                    "Inconsistent input size array %s = %s"
                    " for the design variable of length %s"
                    % (key, x_vect.shape, var_length)
                )
            if key in masking_data_names:
                x_mask[i_min:i_max] = True
            i_min = i_max

        return x_vect[x_mask]

    def unmask_x(
        self,
        masking_data_names,  # type: Iterable[str]
        x_masked,  # type: ndarray
        all_data_names=None,  # type: Optional[Iterable[str]]
        x_full=None,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Unmask a vector from a subset of names, with respect to a set of names.

        Args:
            masking_data_names: The names of the kept data.
            x_masked: The boolean vector to unmask.
            all_data_names: The set of all names.
                If None, use the design variables stored in the design space.
            x_full: The default values for the full vector.
                If None, use the zero vector.

        Returns:
            The vector related to the input mask.
        """
        return self.unmask_x_swap_order(
            masking_data_names, x_masked, all_data_names, x_full
        )

    def _get_dv_length(
        self,
        variable_name,  # type: str
    ):  # type: (...) -> int
        """Retrieve the length of a variable.

        This method relies on the size declared in the design space.

        Args:
            variable_name: The name of the variable.

        Returns:
            The size of the variable.
        """
        return self.opt_problem.design_space.get_size(variable_name)

    def _get_x_mask_swap(
        self,
        masking_data_names,  # type: Iterable[str]
        all_data_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> Tuple[Dict[str,Tuple[int,int]],int,int]
        """Get a mask from a subset of names, with respect to a set of names.

        This method eventually swaps the order of the values
        if the order of the data names is inconsistent between these sets.

        Args:
            masking_data_names: The names of data to keep.
            all_data_names: The set of all names.
                If None, use the design variables stored in the design space.

        Returns:
            A mask as well as
            the dimension of the restricted variable space
            and the dimension of the original variable space.

            The mask is a dictionary
            indexed by the names of the variables coming from the subset.
            For a given name,
            the value is a tuple
            whose first component is its lowest dimension in the original space
            and the second one is the lowest dimension of the variable that follows it
            in the original space.
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        i_min = 0
        x_values_dict = {}
        n_x = 0
        for key in all_data_names:
            i_max = i_min + self._get_dv_length(key)
            if key in masking_data_names:
                x_values_dict[key] = (i_min, i_max)
                n_x += i_max - i_min
            i_min = i_max
        return x_values_dict, n_x, i_max

    def unmask_x_swap_order(
        self,
        masking_data_names,  # type: Iterable[str]
        x_masked,  # type: ndarray
        all_data_names=None,  # type: Optional[Iterable[str]]
        x_full=None,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Unmask a vector from a subset of names, with respect to a set of names.

        This method eventually swaps the order of the values
        if the order of the data names is inconsistent between these sets.

        Args:
            masking_data_names: The names of the kept data.
            x_masked: The boolean vector to unmask.
            all_data_names: The set of all names.
                If None, use the design variables stored in the design space.
            x_full: The default values for the full vector.
                If None, use the zero vector.

        Returns:
            The vector related to the input mask.
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        x_values_dict, _, len_x = self._get_x_mask_swap(
            masking_data_names, all_data_names
        )
        if x_full is None:
            x_unmask = zeros(len_x, dtype=x_masked.dtype)
        else:
            x_unmask = copy(x_full)

        i_x = 0
        for key in all_data_names:
            if key in x_values_dict:
                i_min, i_max = x_values_dict[key]
                n_x = i_max - i_min
                if x_masked.size < i_x + n_x:
                    raise ValueError(
                        "Inconsistent data shapes !\n"
                        "Try to unmask data %s of length %s\n"
                        "With values of length: %s" % (key, n_x, x_masked.size)
                    )
                x_unmask[i_min:i_max] = x_masked[i_x : i_x + n_x]
                i_x += n_x
        return x_unmask

    def mask_x_swap_order(
        self,
        masking_data_names,  # type: Iterable[str]
        x_vect,  # type: ndarray
        all_data_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> ndarray
        """Mask a vector from a subset of names, with respect to a set of names.

        This method eventually swaps the order of the values
        if the order of the data names is inconsistent between these sets.

        Args:
            masking_data_names: The names of the kept data.
            x_vect: The vector to mask.
            all_data_names: The set of all names.
                If None, use the design variables stored in the design space.

        Returns:
            The masked version of the input vector.
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        x_values_dict, n_x, _ = self._get_x_mask_swap(
            masking_data_names, all_data_names
        )
        x_masked = zeros(n_x, dtype=x_vect.dtype)
        i_max = 0
        i_min = 0
        for key in masking_data_names:
            if key not in x_values_dict:
                raise ValueError(
                    "Inconsistent inputs of masking. "
                    "Key %s is in masking_data_names %s "
                    "but not in provided all_data_names : %s!"
                    % (key, masking_data_names, all_data_names)
                )
            value = x_values_dict[key]
            i_max += value[1] - value[0]
            len_x = len(x_vect)
            if len(x_masked) < i_max or len_x <= value[0] or len_x < value[1]:
                raise ValueError(
                    "Inconsistent input array size of values array %s "
                    "with reference data shape %s, "
                    "for data named: %s of size: %s"
                    % (x_vect, x_vect.shape, key, i_max)
                )
            x_masked[i_min:i_max] = x_vect[value[0] : value[1]]
            i_min = i_max

        return x_masked

    def _remove_unused_variables(self):  # type: (...) -> None
        """Remove variables in the design space that are not discipline inputs."""
        design_space = self.opt_problem.design_space
        disciplines = self.get_top_level_disc()
        all_inputs = set(
            var for disc in disciplines for var in disc.get_input_data_names()
        )
        for name in set(design_space.variables_names):
            if name not in all_inputs:
                design_space.remove_variable(name)

    def _remove_sub_scenario_dv_from_ds(self):  # type: (...) -> None
        """Remove the sub scenarios design variables from the design space."""
        for scenario in self.get_sub_scenarios():
            loc_vars = scenario.design_space.variables_names
            for var in loc_vars:
                if var in self.design_space.variables_names:
                    self.design_space.remove_variable(var)

    def _build_objective_from_disc(
        self,
        objective_name,  # type: str
        discipline=None,  # type: Optional[MDODiscipline]
        top_level_disc=True,  # type: bool
    ):  # type: (...) -> None
        """Build the objective function from the discipline able to compute it.

        Args:
            objective_name: The name of the objective function.
            discipline: The discipline computing the objective function.
                If None, the discipline is detected from the inner disciplines.
            top_level_disc: If True, search the discipline among the top level ones.
        """
        if isinstance(objective_name, string_types):
            objective_name = [objective_name]
        obj_mdo_fun = self._get_function_from(
            objective_name, discipline, top_level_disc=top_level_disc
        )
        obj_mdo_fun.f_type = MDOFunction.TYPE_OBJ
        self.opt_problem.objective = obj_mdo_fun
        if self._maximize_objective:
            self.opt_problem.change_objective_sign()

    def _get_function_from(
        self,
        output_names,  # type: Sequence[str]
        discipline=None,  # type: Optional[MDODiscipline]
        top_level_disc=True,  # type:bool
        x_names=None,  # type: Optional[Sequence[str]]
        all_data_names=None,  # type:Optional[Iterable[str]]
        differentiable=True,  # type: bool
    ):  # type: (...) -> MDOFunction
        """Build a function able to compute various outputs.

        Args:
            output_names: The names of the outputs.
            discipline: The discipline computing these outputs.
                If None, the discipline is detected from the inner disciplines.
            top_level_disc: If True, search the discipline among the top level ones.
            x_names: The names of the design variables.
                If None, use self.get_x_names_of_disc(discipline).
            all_data_names: The reference data names for masking x.
                If None, use self.get_optim_variables_names().
            differentiable: If True, then inputs and outputs are added
                to the list of variables to be differentiated.

        Returns:
            The function able to compute the values of the given outputs.
        """
        if discipline is None:
            gen = self._get_generator_from(output_names, top_level_disc=top_level_disc)
            discipline = gen.discipline
        else:
            gen = MDOFunctionGenerator(discipline)

        if x_names is None:
            x_names = self.get_x_names_of_disc(discipline)

        out_x_func = gen.get_function(
            x_names, output_names, differentiable=differentiable
        )

        def func(
            x_vect,  # type: ndarray
        ):  # type: (...) -> ndarray
            """Function to compute the outputs.

            Args:
                x_vect: The design variable vector.

            Returns:
                The value of the outputs.
            """
            x_of_disc = self.mask_x_swap_order(x_names, x_vect, all_data_names)
            obj_allx_val = out_x_func(x_of_disc)
            return obj_allx_val

        masked_func = MDOFunction(
            func,
            out_x_func.name,
            f_type=MDOFunction.TYPE_OBJ,
            args=x_names,
            expr=out_x_func.expr,
            dim=out_x_func.dim,
            outvars=out_x_func.outvars,
        )
        if out_x_func.has_jac():

            def func_jac(
                x_vect,  # type: ndarray
            ):  # type: (...) -> ndarray
                """Function to compute the gradient of the outputs.

                Args:
                    x_vect: The design variable vector.

                Returns:
                    The value of the gradient of the outputs.
                """
                x_of_disc = self.mask_x_swap_order(x_names, x_vect, all_data_names)

                loc_jac = out_x_func.jac(x_of_disc)  # pylint: disable=E1102

                if len(loc_jac.shape) == 1:
                    # This is surprising but there is a duality between the
                    # masking operation in the function inputs and the
                    # unmasking of its outputs
                    jac = self.unmask_x_swap_order(x_names, loc_jac, all_data_names)
                else:
                    n_outs = loc_jac.shape[0]
                    jac = empty((n_outs, x_vect.size), dtype=x_vect.dtype)
                    for func_ind in range(n_outs):
                        gr_u = self.unmask_x_swap_order(
                            x_names, loc_jac[func_ind, :], all_data_names
                        )
                        jac[func_ind, :] = gr_u
                return jac

            masked_func.jac = func_jac

        return masked_func

    def get_optim_variables_names(self):  # type: (...) -> List[str]
        """Get the optimization unknown names to be provided to the optimizer.

        This is different from the design variable names provided by the user,
        since it depends on the formulation,
        and can include target values for coupling for instance in IDF.

        Returns:
            The optimization variable names.
        """
        return self.opt_problem.design_space.variables_names

    def get_x_names_of_disc(
        self,
        discipline,  # type: MDODiscipline
    ):  # type: (...) -> List[str]
        """Get the design variables names of a given discipline.

        Args:
            discipline: The discipline.

        Returns:
             The names of the design variables.
        """
        optim_variables_names = self.get_optim_variables_names()
        input_names = discipline.get_input_data_names()
        return [name for name in optim_variables_names if name in input_names]

    def get_sub_disciplines(self):  # type: (...) ->List[MDODiscipline]
        """Accessor to the sub-disciplines.

        This method lists the sub scenarios' disciplines.

        Returns:
            The sub-disciplines.
        """
        sub_disc = []

        def add_to_sub(
            disc_list,  # type:Iterable[MDODiscipline]
        ):  # type: (...) -> None
            """Add the disciplines of the sub-scenarios if not already added it.

            Args:
                disc_list: The disciplines.
            """
            for disc in disc_list:
                if disc not in sub_disc:
                    sub_disc.append(disc)

        for discipline in self.disciplines:
            if hasattr(discipline, "disciplines"):
                add_to_sub(discipline.disciplines)
            else:
                add_to_sub([discipline])

        return sub_disc

    def get_sub_scenarios(self):  # type: (...) -> List[Scenario]
        """List the disciplines that are actually scenarios.

        Returns:
            The scenarios.
        """
        return [disc for disc in self.disciplines if disc.is_scenario()]

    def _set_defaultinputs_from_ds(self):  # type: (...) -> None
        """Initialize the top level disciplines from the design space."""
        if not self.opt_problem.design_space.has_current_x():
            return
        x_dict = self.opt_problem.design_space.get_current_x_dict()
        for disc in self.get_top_level_disc():
            inputs = disc.get_input_data_names()
            curr_x_disc = {name: x_dict[name] for name in inputs if name in x_dict}
            disc.default_inputs.update(curr_x_disc)

    def get_expected_workflow(
        self,
    ):  # type: (...) -> List[ExecutionSequence,Tuple[ExecutionSequence]]
        """Get the expected sequence of execution of the disciplines.

        This method is used for the XDSM representation
        and can be overloaded by subclasses.

        For instance:

        * [A, B] denotes the execution of A,
          then the execution of B
        * (A, B) denotes the concurrent execution of A and B
        * [A, (B, C), D] denotes the execution of A,
          then the concurrent execution of B and C,
          then the execution of D.

        Returns:
            A sequence of elements which are either
            an :class:`.ExecutionSequence`
            or a tuple of :class:`.ExecutionSequence` for concurrent execution.
        """
        raise NotImplementedError()

    def get_expected_dataflow(
        self,
    ):  # type: (...) -> List[Tuple[MDODiscipline,MDODiscipline,List[str]]]
        """Get the expected data exchange sequence.

        This method is used for the XDSM representation
        and can be overloaded by subclasses.

        Returns:
            The expected sequence of data exchange
            where the i-th item is described by the starting discipline,
            the ending discipline and the coupling variables.
        """
        raise NotImplementedError()

    @classmethod
    def get_default_sub_options_values(
        cls, **options  # type:str
    ):  # type: (...) -> Dict  # pylint: disable=W0613
        """Get the default values of the sub-options of the formulation.

        When some options of the formulation depend on higher level options,
        the default values of these sub-options may be obtained here,
        mainly for use in the API.

        Args:
            **options: The options required to deduce the sub-options grammar.

        Returns:
            Either None or the sub-options default values.
        """

    @classmethod
    def get_sub_options_grammar(
        cls, **options  # type: str
    ):  # type: (...) -> JSONGrammar # pylint: disable=W0613
        """Get the sub-options grammar.

        When some options of the formulation depend on higher level options,
        the schema of the sub-options may be obtained here,
        mainly for use in the API.

        Args:
            **options: The options required to deduce the sub-options grammar.

        Returns:
            Either None or the sub-options grammar.
        """



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The F1 to measure the quality of a classification algorithm.

The F1 is defined by

.. math::

    F_1 = 2\\frac{\\mathit{precision}\\mathit{recall}}
        {\\mathit{precision}+\\mathit{recall}}

where
:math:`\\mathit{precision}` is the number of correctly predicted positives
divided by the total number of *predicted* positives
and :math:`\\mathit{recall}` is the number of correctly predicted positives
divided by the total number of *true* positives.
"""
from __future__ import division, unicode_literals

from typing import Union

from numpy import ndarray
from sklearn.metrics import f1_score

from gemseo.mlearning.classification.classification import MLClassificationAlgo
from gemseo.mlearning.qual_measure.error_measure import MLErrorMeasure


class F1Measure(MLErrorMeasure):
    """The F1 measure for machine learning."""

    SMALLER_IS_BETTER = False

    def __init__(
        self,
        algo,  # type: MLClassificationAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for classification.
        """
        super(F1Measure, self).__init__(algo)

    def _compute_measure(
        self,
        outputs,  # type: ndarray
        predictions,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        if multioutput:
            raise NotImplementedError("F1 is only defined for single target.")
        return f1_score(outputs, predictions, average="weighted")
